import networkx
from pymead.core.bezier import Bezier

from pymead.core.constraints import *
from pymead.core.constraint_equations import *
from pymead.core.line import PolyLine
from pymead.core.point import Point


class GCS(networkx.DiGraph):
    """
    A Geometric Constraint Solver (GCS) used to maintain constraints between points in pymead, implemented using
    the directed graph class in ``networkx``. Instances of this
    class should not normally be created directly; the geometry collection creates objects of this class when
    ``GeometryCollection.add_constraint`` is called.

    This constraint solver, while not powerful enough to handle arbitrary systems of equations, is nevertheless
    sufficient for handling many common types of constraints needed for airfoil systems. The constraints are solved
    using a graph-constructive approach and maintained using simple rigid body transformations. These transformations
    occur for a given set of points when the parameter value of a constraint "upstream" of this set of points is
    modified. All the points downstream of this constraint are handled as a rigid body (even if they are not
    fully constrained as such) to preserve each of the relative constraints.

    The primary constraints in pymead are distance constraints and relative angle constraints (with specialized options
    for perpendicular and antiparallel constraints). The angle constraints are formed between three points
    (start, vertex, and end). Two other special kinds of constraints, symmetry constraints and radius of curvature
    constraints, are solved following the rigid body transformation.
    """
    def __init__(self):
        """
        Constructor for the geometric constraint solver.
        """
        super().__init__()
        self.points = {}
        self.roots = []
        self.cluster_angle_params = {}
        self.geo_col = None
        self.partial_symmetry_solves = []

    def add_point(self, point: Point):
        """
        Adds a point (node) to the graph (and to the ``"points"`` attribute). Also assigns this graph to the point.

        Parameters
        ----------
        point: Point
            Point to add

        Returns
        -------

        """
        point.gcs = self
        self.add_node(point)
        self.points[point.name()] = point

    def remove_point(self, point: Point):
        """
        Removes a point (node) from the graph (and from the ``"points"`` attribute).

        Parameters
        ----------
        point: Point
            Point to remove

        Returns
        -------

        """
        self.remove_node(point)
        self.points.pop(point.name())

    def _check_if_constraint_creates_new_cluster(self, constraint: GeoCon):
        """
        Determines if a new cluster is formed by adding this constraint by checking if there are any edges connected
        in either direction to any of the points associated with the constraint. Only if no edges are found is a new
        cluster created.

        Parameters
        ----------
        constraint: GeoCon
            Constraint to analyze

        Returns
        -------
        bool
            ``True`` if adding this constraint forms a cluster, ``False`` otherwise

        """
        for point in constraint.child_nodes:
            # If there is already an edge attached to any of the points in this constraint, do not create a new root
            if len(self.in_edges(nbunch=point)) > 0 or len(self.out_edges(nbunch=point)) > 0:
                return False
        return True

    def _set_edge_as_root(self, u: Point, v: Point):
        """
        Sets the given edge as a root by setting the ``u`` value as the root and the ``v`` value as the rotation
        handle of the constraint cluster.

        Parameters
        ----------
        u: Point
            Starting node of the edge

        v: Point
            Terminating node of the edge

        Returns
        -------

        """
        u.root = True
        v.rotation_handle = True
        if u not in [edge[0] for edge in self.roots]:
            self.roots.append((u, v))

        cluster_angle_exceptions = [
            # Do not add a cluster angle parameter if the root was created as part of an antiparallel constraint between
            # a polyline and another curve
            any([isinstance(curve, PolyLine) for curve in u.curves]) and not all(
                [u in curve.point_sequence().points() for curve in u.curves]),

            # Do not add a cluster angle parameter if the root was created as part of an antiparallel constraint between
            # a curve defined by symmetric constraints and a normal Bézier curve
            (SymmetryConstraint.check_if_point_is_symmetric_target(u) and
                SymmetryConstraint.check_if_point_is_symmetric_target(v))
        ]
        if any(cluster_angle_exceptions):
            v.rotation_handle = False  # TODO: check if this change works for other cases
            return

        if v.rotation_param is not None:
            param = v.rotation_param
        else:
            param = self.geo_col.add_param(value=u.measure_angle(v), name="ClusterAngle-1", unit_type="angle", root=u,
                                           rotation_handle=v)
        self.cluster_angle_params[u] = param
        param.gcs = self

    def _delete_root_status(self, root_node: Point, rotation_handle_node: Point = None):
        """
        Removes root status from the given edge.

        Parameters
        ----------
        root_node: Point
            ``u``-value of the edge

        rotation_handle_node: Point
            ``v``-value of the edge

        Returns
        -------

        """
        root_node.root = False
        if rotation_handle_node is not None:
            rotation_handle_node.rotation_handle = False
            rotation_handle_node.rotation_param = None
        if root_node in self.cluster_angle_params:
            cluster_angle_param = self.cluster_angle_params.pop(root_node)
            if cluster_angle_param is not None:
                self.geo_col.remove_pymead_obj(cluster_angle_param, constraint_removal=True)
        root_idx = [r[0] for r in self.roots].index(root_node)
        self.roots.pop(root_idx)

    def _identify_root_from_rotation_handle(self, rotation_handle: Point) -> Point:
        """
        Computes the root of the cluster given the rotation handle as the ``u``-value of the edge incident to the
        rotation handle.

        Parameters
        ----------
        rotation_handle: Point
            Rotation handle of the constraint cluster

        Returns
        -------
        Point
            Root of the constraint cluster

        """
        if not isinstance(rotation_handle, Point):
            raise ValueError(f"Detected rotation handle that is not a point ({rotation_handle = })")
        in_edges = [edge for edge in self.in_edges(nbunch=rotation_handle)]
        if not len(in_edges) == 1:
            raise ValueError("Invalid rotation handle. Rotation should have exactly one incident edge "
                             "(the cluster root)")
        return in_edges[0][0]

    def _identify_rotation_handle(self, root_node: Point):
        """
        Identifies the rotation handle by starting at the root node and testing each of the root node's neighbors
        until the rotation handle is found.

        Parameters
        ----------
        root_node: Point
            Root of the constraint cluster

        Returns
        -------
        Point
            Rotation handle

        """
        for edge in self.out_edges(nbunch=root_node):
            if edge[1].rotation_handle:
                return edge[1]
        raise ValueError("Could not identify rotation handle")

    def _identify_and_delete_root(self, root_node: Point):
        """
        Identifies the root edge by starting at the root node and testing each of the root node's neighbors
        until the rotation handle is found (the ``v``-value of the root edge). Then, ``self._delete_root_status`` is
        applied to demote the edge.

        Parameters
        ----------
        root_node: Point
            Root of the constraint cluster

        Returns
        -------

        """
        for edge in self.out_edges(nbunch=root_node):
            if edge[1].rotation_handle:
                self._delete_root_status(root_node, edge[1])
                return
        self._delete_root_status(root_node=root_node)

    def _get_unique_roots_from_constraint(self, constraint: GeoCon) -> typing.List[Point]:
        """
        Gets the unique roots associated with the addition of the input constraint. Used to determine whether a cluster
        merge should occur.

        Parameters
        ----------
        constraint: GeoCon
            Constraint from which to determine the unique, connected set of cluster roots

        Returns
        -------
        typing.List[Point]
            List of points representing the unique roots found. An error is raised if more than two are found, since
            this should not be possible when adding a constraint.

        """
        unique_roots = []
        for point in constraint.child_nodes:
            root = self._discover_root_from_node(point)
            if root and root not in unique_roots:
                unique_roots.append(root)

        if len(unique_roots) > 2:
            raise ValueError("Found more than two unique roots connected to the constraint being added")

        return unique_roots

    def _discover_root_from_node(self, source_node: Point):
        """
        Traverse backward along the constraint graph from the ``source_node`` until the root node is reached.
        Returns ``None`` if the root was not found.

        Parameters
        ----------
        source_node: Point
            Starting point

        Returns
        -------
        Point or None
            The root node/point if found, otherwise ``None``
        """
        for node in networkx.bfs_tree(self, source=source_node, reverse=True):
            if node.root:
                return node
        for node in networkx.bfs_tree(self, source=source_node, reverse=False):
            if node.root:
                return node
        return None

    def _orient_flow_away_from_root(self, root: Point) -> typing.List[GeoCon]:
        """
        Orients the flow of the edges away from the given root of a constraint cluster. Critical method that ensures
        the rigid body transformation triggered by a given constraint value change applies to the correct set of points
        to preserve the constraint state.

        Parameters
        ----------
        root: Point
            Root of the constraint cluster

        Returns
        -------
        typing.List[GeoCon]
            A list of constraints associated with the edges that were affected by this reorientation. These constraints
            need re-assigning to apply the constraint data to the flipped edges.
        """
        constraints_needing_reassign = []
        while True:
            constraints_to_flip = []
            points_from_root = [point for point in networkx.dfs_preorder_nodes(self, source=root)]
            for point in points_from_root:
                in_edges = self.in_edges(nbunch=point)
                u_values = [in_edge[0] for in_edge in in_edges]
                v_values = [in_edge[1] for in_edge in in_edges]
                set_difference = list(set(u_values) - set(points_from_root))
                for u_value in set_difference:
                    constraints_to_flip.extend(u_value.geo_cons)
                    in_edge_idx = u_values.index(u_value)
                    v = v_values[in_edge_idx]
                    self.remove_edge(u_value, v)
                    self.add_edge(v, u_value)
            constraints_needing_reassign.extend(constraints_to_flip)
            if len(constraints_to_flip) == 0:
                break
        return list(set(constraints_needing_reassign))

    def _check_if_root_flows_into_polyline(self, root_node: Point):
        for point in networkx.dfs_preorder_nodes(self, source=root_node):
            if any([isinstance(curve, PolyLine) and point not in curve.point_sequence().points()
                    for curve in point.curves]):
                # This is the case where the root is the newly created tangent point on the polyline
                return False
            if point is root_node:
                continue
            if any([isinstance(curve, PolyLine) for curve in point.curves]):
                return point
        return False

    def move_root(self, new_root: Point):
        old_root = self._discover_root_from_node(new_root)
        self._identify_and_delete_root(old_root)
        needs_set_edge = False
        for nbr in self.adj[new_root]:
            self._set_edge_as_root(new_root, nbr)
            break
        else:
            needs_set_edge = True
        constraints_needing_reassign = self._orient_flow_away_from_root(new_root)
        self._reassign_constraints(constraints_needing_reassign)
        if not needs_set_edge:
            return

        for nbr in self.adj[new_root]:
            self._set_edge_as_root(new_root, nbr)
            break
        else:
            raise ValueError("Failed to move root")

    def _reassign_distance_constraint(self, dist_con: DistanceConstraint):
        """
        Re-assigns distance constraint data to a flipped edge.

        Parameters
        ----------
        dist_con: DistanceConstraint
            Constraint to re-assign

        Returns
        -------

        """
        edge_data_12 = self.get_edge_data(dist_con.p1, dist_con.p2)
        if edge_data_12 is not None:
            networkx.set_edge_attributes(self, {(dist_con.p1, dist_con.p2): dist_con}, name="distance")
            return
        edge_data_21 = self.get_edge_data(dist_con.p2, dist_con.p1)
        if edge_data_21 is not None:
            networkx.set_edge_attributes(self, {(dist_con.p2, dist_con.p1): dist_con}, name="distance")
            return
        raise ValueError(f"Could not reassign distance constraint {dist_con}")

    def _reassign_angle_constraint(self, angle_con: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint):
        """
        Re-assigns angle constraint data to a swapped edge.

        Parameters
        ----------
        angle_con: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint
            Constraint to re-assign

        Returns
        -------

        """
        edge_data_21 = self.get_edge_data(angle_con.p2, angle_con.p1)
        if edge_data_21 is not None and "angle" not in edge_data_21.keys():
            networkx.set_edge_attributes(self, {(angle_con.p2, angle_con.p1): angle_con}, name="angle")
            return
        edge_data_23 = self.get_edge_data(angle_con.p2, angle_con.p3)
        if edge_data_23 is not None and "angle" not in edge_data_23.keys():
            networkx.set_edge_attributes(self, {(angle_con.p2, angle_con.p3): angle_con}, name="angle")
            return

    def _reassign_constraint(self, constraint: GeoCon):
        """
        Convenience method that applies either ``self._reassign_distance_constraint`` or
        ``self._reassign_angle_constraint`` depending on the type of the input constraint.

        Parameters
        ----------
        constraint: GeoCon
            Constraint to re-assign

        Returns
        -------

        """
        if isinstance(constraint, DistanceConstraint):
            self._reassign_distance_constraint(constraint)
        elif isinstance(constraint, RelAngle3Constraint) or isinstance(constraint, Perp3Constraint) or isinstance(
                constraint, AntiParallel3Constraint):
            self._reassign_angle_constraint(constraint)
        else:
            raise ValueError(f"Constraint reassignment for constraints of type {constraint} are not implemented")

    def _reassign_constraints(self, constraints: typing.List[GeoCon]):
        """
        Convenience method that applies ``self._reassign_constraint`` to a list of constraints

        Parameters
        ----------
        constraints: typing.List[GeoCon]
            List of constraints to re-assign

        Returns
        -------

        """
        for constraint in constraints:
            if (isinstance(constraint, DistanceConstraint) or isinstance(constraint, RelAngle3Constraint) or
                    isinstance(constraint, Perp3Constraint) or isinstance(constraint, AntiParallel3Constraint)):
                self._reassign_constraint(constraint)

    def _check_distance_constraint_for_duplicates(self, dist_con: DistanceConstraint):
        for gc in self.geo_col.container()["geocon"].values():
            if gc is dist_con:
                continue
            if not isinstance(gc, DistanceConstraint):
                continue
            if (dist_con.p1 is gc.p1 and dist_con.p2 is gc.p2) or (dist_con.p1 is gc.p2 and dist_con.p2 is gc.p1):
                raise ValueError("A distance constraint already exists between these two points")

    def _check_angle_constraint_for_duplicates(self, angle_con: RelAngle3Constraint or AntiParallel3Constraint or
                                                                Perp3Constraint):
        for gc in self.geo_col.container()["geocon"].values():
            if gc is angle_con:
                continue
            if not isinstance(gc, RelAngle3Constraint) and not isinstance(
                    gc, AntiParallel3Constraint) and not isinstance(gc, Perp3Constraint):
                continue
            if (gc.p1 is angle_con.p1 and gc.p3 is angle_con.p3) or (
                    gc.p1 is angle_con.p3 and gc.p3 is angle_con.p1):
                raise ValueError("An angle constraint already exists between these three points")

    def check_constraint_for_duplicates(self, geo_con: GeoCon):
        if isinstance(geo_con, DistanceConstraint):
            self._check_distance_constraint_for_duplicates(geo_con)
        elif isinstance(geo_con, RelAngle3Constraint) or isinstance(
                geo_con, AntiParallel3Constraint) or isinstance(geo_con, Perp3Constraint):
            self._check_angle_constraint_for_duplicates(geo_con)

    def _assign_distance_constraint(self, dist_con: DistanceConstraint):
        """
        Assigns a distance constraint by first adding an edge from ``p1`` to ``p2`` with no data and subsequently
        applying the constraint data.

        Parameters
        ----------
        dist_con: DistanceConstraint
            Constraint to assign

        Returns
        -------

        """
        if (SymmetryConstraint.check_if_point_is_symmetric_target(dist_con.p1) and
            SymmetryConstraint.check_if_point_is_symmetric_target(dist_con.p2)):
            raise ValueError("Could not assign distance constraint")
        if SymmetryConstraint.check_if_point_is_symmetric_target(dist_con.p1):
            self.add_edge(dist_con.p1, dist_con.p2)
        elif SymmetryConstraint.check_if_point_is_symmetric_target(dist_con.p2):
            self.add_edge(dist_con.p2, dist_con.p1)
        else:
            self.add_edge(dist_con.p1, dist_con.p2)
        self._reassign_distance_constraint(dist_con)

    def _assign_angle_constraint(self, angle_con: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint):
        """
        Assigns an angle constraint by first adding an edge from either ``p2`` to ``p1`` or ``p2`` to ``p3``
        (depending on which edge already has data) with no data and subsequently applying the constraint data.


        Parameters
        ----------
        angle_con: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint
            Constraint to assign

        Returns
        -------

        """
        if (SymmetryConstraint.check_if_point_is_symmetric_target(angle_con.p1) and
                SymmetryConstraint.check_if_point_is_symmetric_target(angle_con.p2) and
                SymmetryConstraint.check_if_point_is_symmetric_target(angle_con.p3)):
            raise ValueError("Could not add angle constraint")
        if (not (self.get_edge_data(angle_con.p1, angle_con.p2)
                or self.get_edge_data(angle_con.p2, angle_con.p1)) and
                not (SymmetryConstraint.check_if_point_is_symmetric_target(angle_con.p1) and
                     SymmetryConstraint.check_if_point_is_symmetric_target(angle_con.p2))):
            self.add_edge(angle_con.p2, angle_con.p1)
        if (not (self.get_edge_data(angle_con.p2, angle_con.p3) or
                 self.get_edge_data(angle_con.p3, angle_con.p2)) and
                  not (SymmetryConstraint.check_if_point_is_symmetric_target(angle_con.p2) and
                       SymmetryConstraint.check_if_point_is_symmetric_target(angle_con.p3))):
            self.add_edge(angle_con.p2, angle_con.p3)
        self._reassign_angle_constraint(angle_con)

    def _add_ghost_edges_to_angle_constraint(
            self, constraint: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint):
        """
        Adds any required "ghost" edges (edges with no data) to an angle constraint to ensure that there are edges
        connecting both (``p1`` and ``p2``) and (``p2`` and ``p3``). Performing this action ensures that any upstream
        changes in constraint parameter value properly flow down to this angle constraint.

        Parameters
        ----------
        constraint: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint
            Angle constraint to add ghost edges to (if necessary)

        Returns
        -------

        """
        # Only add ghost edges if there is not a concrete edge present (even if facing the wrong direction)
        if not (self.get_edge_data(constraint.p1, constraint.p2)
                or self.get_edge_data(constraint.p2, constraint.p1)):
            if (SymmetryConstraint.check_if_point_is_symmetric_target(constraint.p1) and
                SymmetryConstraint.check_if_point_is_symmetric_target(constraint.p2)):
                self.add_edge(constraint.p1, constraint.p2)
                if constraint.p2.root:
                    self.move_root(constraint.p1)
            else:
                self.add_edge(constraint.p2, constraint.p1)
        if not (self.get_edge_data(constraint.p2, constraint.p3) or
                self.get_edge_data(constraint.p3, constraint.p2)):
            if (SymmetryConstraint.check_if_point_is_symmetric_target(constraint.p2) and
                    SymmetryConstraint.check_if_point_is_symmetric_target(constraint.p3)):
                self.add_edge(constraint.p3, constraint.p2)
                if constraint.p2.root:
                    self.move_root(constraint.p3)
            else:
                self.add_edge(constraint.p2, constraint.p3)

    # def _assign_symmetry_constraint(self, constraint: SymmetryConstraint):
    #     self.add_edge(constraint.p1, constraint.p4)
    #     self.add_edge(constraint.p2, constraint.p4)
    #     self.add_edge(constraint.p3, constraint.p4)

    def _test_if_cluster_is_branching(self, root_node: Point) -> bool:
        """
        Computes the subgraph corresponding to the constraint cluster given by the input root node, and checks if
        this subgraph is branching: each node must have exactly one parent.

        Parameters
        ----------
        root_node: Point
            Root of the cluster to analyze

        Returns
        -------
        bool
            Whether this constraint cluster is branching. If not, an error should be thrown (since closed loops
            of constraints are not possible in pymead)

        """
        subgraph = self.subgraph([node for node in networkx.dfs_preorder_nodes(self, source=root_node)])
        return networkx.is_branching(subgraph)

    def _merge_clusters_with_constraint(self, unique_roots: typing.List[Point]):

        def _determine_merged_cluster_root():
            current_roots = [r[0] for r in self.roots]
            if current_roots.index(unique_roots[0]) < current_roots.index(unique_roots[1]):
                return unique_roots[0]
            else:
                return unique_roots[1]

        def _delete_root_status_of_other_roots(new_root: Point):
            for root in unique_roots:
                if root is new_root:
                    continue
                self._identify_and_delete_root(root)
                break

        merged_cluster_root = _determine_merged_cluster_root()
        _delete_root_status_of_other_roots(merged_cluster_root)
        constraints_needing_reassign = self._orient_flow_away_from_root(merged_cluster_root)
        self._reassign_constraints(constraints_needing_reassign)
        return merged_cluster_root

    def add_constraint(self, constraint: GeoCon):

        # Check if this constraint creates a new cluster
        first_constraint_in_cluster = self._check_if_constraint_creates_new_cluster(constraint)

        # If it does, set it as the root constraint
        if first_constraint_in_cluster and (
                isinstance(constraint, DistanceConstraint) or
                isinstance(constraint, RelAngle3Constraint) or
                isinstance(constraint, Perp3Constraint) or
                isinstance(constraint, AntiParallel3Constraint)):
            self._set_edge_as_root(constraint.p1, constraint.p2)

        # If the constraint has a Param associated with it, pass the GCS reference to this parameter
        if constraint.param() is not None:
            constraint.param().gcs = self

        if isinstance(constraint, DistanceConstraint):
            self._assign_distance_constraint(constraint)
        elif isinstance(constraint, RelAngle3Constraint) or isinstance(
                constraint, AntiParallel3Constraint) or isinstance(constraint, Perp3Constraint):
            self._assign_angle_constraint(constraint)
            self._add_ghost_edges_to_angle_constraint(constraint)
        # elif isinstance(constraint, SymmetryConstraint):
        #     self._assign_symmetry_constraint(constraint)

        if isinstance(constraint, DistanceConstraint) or isinstance(constraint, AntiParallel3Constraint) or isinstance(
                constraint, Perp3Constraint) or isinstance(constraint, RelAngle3Constraint):
            unique_roots = self._get_unique_roots_from_constraint(constraint)
            merge_clusters = False if len(unique_roots) < 2 else True
            if merge_clusters:
                root = self._merge_clusters_with_constraint(unique_roots)
            else:
                constraints_to_reassign = self._orient_flow_away_from_root(unique_roots[0])
                self._reassign_constraints(constraints_to_reassign)
                root = self._discover_root_from_node(constraint.p1)

            # Check if the addition of this constraint creates a closed loop
            is_branching = self._test_if_cluster_is_branching(root)
            if not is_branching:
                raise ValueError("Detected a closed loop in the constraint graph. Closed loop sets of constraints "
                                 "are currently not supported in pymead")

            new_root = self._check_if_root_flows_into_polyline(root)
            if new_root:
                self.move_root(new_root)

        elif isinstance(constraint, SymmetryConstraint):
            points_solved = self.solve_symmetry_constraint(constraint)
            self.update_canvas_items(points_solved)
        elif isinstance(constraint, ROCurvatureConstraint):
            points_solved = self.solve_roc_constraint(constraint)
            self.update_canvas_items(points_solved)

    def _remove_distance_constraint_from_directed_edge(self, constraint: DistanceConstraint):
        edges_removed = None
        edge_data_12 = self.get_edge_data(constraint.p1, constraint.p2)
        if edge_data_12 is not None and "distance" in edge_data_12.keys() and edge_data_12["distance"] is constraint:
            angle_constraint_present = False
            for geo_con in constraint.p2.geo_cons:
                if (isinstance(geo_con, RelAngle3Constraint) or
                        isinstance(geo_con, AntiParallel3Constraint) or
                        isinstance(geo_con, Perp3Constraint) and geo_con.p2 is constraint.p2):
                    angle_constraint_present = True
                    break
            if angle_constraint_present:
                edge_data_12.pop("distance")
            else:
                self.remove_edge(constraint.p1, constraint.p2)
                edges_removed = [(constraint.p1, constraint.p2)]

            return edges_removed

        edge_data_21 = self.get_edge_data(constraint.p2, constraint.p1)
        if edge_data_21 is not None and "distance" in edge_data_21.keys() and edge_data_21["distance"] is constraint:
            angle_constraint_present = False
            for geo_con in constraint.p1.geo_cons:
                if (isinstance(geo_con, RelAngle3Constraint) or
                        isinstance(geo_con, AntiParallel3Constraint) or
                        isinstance(geo_con, Perp3Constraint) and geo_con.p2 is constraint.p1):
                    angle_constraint_present = True
                    break
            if angle_constraint_present:
                edge_data_21.pop("distance")
            else:
                self.remove_edge(constraint.p2, constraint.p1)
                edges_removed = [(constraint.p2, constraint.p1)]

            return edges_removed

    def _remove_angle_constraint_from_directed_edge(
            self, constraint: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint):
        edges_removed = []
        edge_data_21 = self.get_edge_data(constraint.p2, constraint.p1)
        if edge_data_21 is not None and "angle" in edge_data_21.keys() and edge_data_21["angle"] is constraint:
            if "distance" in edge_data_21.keys():
                edge_data_21.pop("angle")
            else:
                self.remove_edge(constraint.p2, constraint.p1)
                edges_removed = [(constraint.p2, constraint.p1)]

            # Remove the ghost edge if there is one
            edge_data_32 = self.get_edge_data(constraint.p3, constraint.p2)
            if edge_data_32 is not None and len(edge_data_32) == 0:
                self.remove_edge(constraint.p3, constraint.p2)
                edges_removed.append((constraint.p3, constraint.p2))

            return edges_removed

        edge_data_23 = self.get_edge_data(constraint.p2, constraint.p3)
        if edge_data_23 is not None and "angle" in edge_data_23.keys() and edge_data_23["angle"] is constraint:
            if "distance" in edge_data_23.keys():
                edge_data_23.pop("angle")
            else:
                self.remove_edge(constraint.p2, constraint.p3)
                edges_removed = [(constraint.p2, constraint.p3)]

            # Remove the ghost edge if there is one
            edge_data_12 = self.get_edge_data(constraint.p1, constraint.p2)
            if edge_data_12 is not None and len(edge_data_12) == 0:
                self.remove_edge(constraint.p1, constraint.p2)
                edges_removed.append((constraint.p1, constraint.p2))

            return edges_removed

    def _assign_new_root_if_required(self, v_of_edge_removed: Point):
        neighbors_of_v = [nbr for nbr in self.adj[v_of_edge_removed]]
        if len(neighbors_of_v) > 0:
            self._set_edge_as_root(v_of_edge_removed, neighbors_of_v[0])
        else:
            pass  # This means we trimmed the end of the branch

    def _update_roots_based_on_constraint_removal(self, edges_removed: typing.List[tuple]):
        for edge_removed in edges_removed:
            if edge_removed[1].rotation_handle:
                self._delete_root_status(edge_removed[0], edge_removed[1])
                for out_edge in self.out_edges(nbunch=edge_removed[0]):
                    self._set_edge_as_root(out_edge[0], out_edge[1])
                    break
            self._assign_new_root_if_required(edge_removed[1])

    def remove_constraint(self, constraint: GeoCon):

        edges_removed = None
        if isinstance(constraint, DistanceConstraint):
            edges_removed = self._remove_distance_constraint_from_directed_edge(constraint)
        elif isinstance(constraint, RelAngle3Constraint) or isinstance(
                constraint, AntiParallel3Constraint) or isinstance(constraint, Perp3Constraint):
            edges_removed = self._remove_angle_constraint_from_directed_edge(constraint)

        if edges_removed is not None:
            self._update_roots_based_on_constraint_removal(edges_removed)

        for child_node in constraint.child_nodes:
            child_node.geo_cons.remove(constraint)

    def _solve_distance_constraint(self, source: DistanceConstraint):
        points_solved = []
        edge_data_p12 = self.get_edge_data(source.p1, source.p2)
        if edge_data_p12 and edge_data_p12["distance"] is source:
            angle = source.p1.measure_angle(source.p2)
            start = source.p2
        else:
            angle = source.p2.measure_angle(source.p1)
            start = source.p1
        old_distance = source.p1.measure_distance(source.p2)
        new_distance = source.param().value()
        dx = (new_distance - old_distance) * np.cos(angle)
        dy = (new_distance - old_distance) * np.sin(angle)

        for point in networkx.bfs_tree(self, source=start):
            point.x().set_value(point.x().value() + dx)
            point.y().set_value(point.y().value() + dy)
            if point not in points_solved:
                points_solved.append(point)

        return points_solved

    def _solve_angle_constraint(self, source: RelAngle3Constraint or AntiParallel3Constraint or Perp3Constraint):

        points_solved = []
        edge_data_p21 = self.get_edge_data(source.p2, source.p1)
        edge_data_p23 = self.get_edge_data(source.p2, source.p3)

        old_angle = (source.p2.measure_angle(source.p1) -
                     source.p2.measure_angle(source.p3)) % (2 * np.pi)
        if isinstance(source, RelAngle3Constraint):
            new_angle = source.param().rad()
        elif isinstance(source, AntiParallel3Constraint):
            new_angle = np.pi
        elif isinstance(source, Perp3Constraint):
            new_angle = np.pi / 2
        else:
            raise ValueError(f"{type(source)} is not a valid type for angle constraint")
        d_angle = new_angle - old_angle
        rotation_point = source.p2

        if edge_data_p21 and "angle" in edge_data_p21 and edge_data_p21["angle"] is source:
            # start = source.p1 if not source.p2.root else source.p2
            # start = source.p2
            start = source.p1
        elif edge_data_p23 and "angle" in edge_data_p23 and edge_data_p23["angle"] is source:
            # start = source.p3 if not source.p2.root else source.p2
            # start = source.p2
            start = source.p3
            d_angle *= -1
        else:
            raise ValueError("Somehow no angle constraint found between the three points")

        rotation_mat = np.array([[np.cos(d_angle), -np.sin(d_angle)], [np.sin(d_angle), np.cos(d_angle)]])
        rotation_point_mat = np.array([[rotation_point.x().value()], [rotation_point.y().value()]])

        # Get all the points that might need to rotate
        additional_branch_starting_points = []
        if start is not source.p2:
            for geo_con in start.geo_cons:
                if geo_con is source:
                    continue
                if not (isinstance(geo_con, AntiParallel3Constraint) or isinstance(
                        geo_con, RelAngle3Constraint) or isinstance(geo_con, Perp3Constraint)):
                    continue
                if geo_con.p2 is not source.p2:
                    continue
                # if not geo_con.p1 in self.adj[source.p2] or not geo_con.p3 in self.adj[source.p2]:
                #     continue
                if start is source.p3 and geo_con.p3 is source.p3:
                    additional_branch_starting_points.append(geo_con.p1)
                elif start is source.p3 and geo_con.p1 is source.p3:
                    additional_branch_starting_points.append(geo_con.p3)
                elif start is source.p1 and geo_con.p3 is source.p1:
                    additional_branch_starting_points.append(geo_con.p1)
                elif start is source.p1 and geo_con.p1 is source.p1:
                    additional_branch_starting_points.append(geo_con.p3)

        all_downstream_points = []
        rotation_handle = None
        # print(f"{start = }, {additional_branch_starting_points = }")
        for source_point in [start, *additional_branch_starting_points]:
            for point in networkx.bfs_tree(self, source=source_point):
                # if point not in points_to_exclude:
                all_downstream_points.append(point)
                if not rotation_handle and point.rotation_handle:
                    rotation_handle = point

        # Get the branch to cut, if there is one
        root_rotation_branch = []
        if source.p2.root and rotation_handle is not None:
            root_rotation_branch = [point for point in networkx.bfs_tree(self, source=rotation_handle)]

        for point in all_downstream_points:
            if point is source.p2 or point in root_rotation_branch:
                continue
            dx_dy = np.array([[point.x().value() - rotation_point.x().value()],
                              [point.y().value() - rotation_point.y().value()]])
            new_xy = (rotation_mat @ dx_dy + rotation_point_mat).flatten()
            point.x().set_value(new_xy[0])
            point.y().set_value(new_xy[1])
            if point not in points_solved:
                points_solved.append(point)
        return points_solved

    def solve(self, source: GeoCon):

        points_solved = []
        symmetry_points_solved = []
        roc_points_solved = []
        if isinstance(source, DistanceConstraint):
            points_solved.extend(self._solve_distance_constraint(source))
        elif (isinstance(source, RelAngle3Constraint) or isinstance(source, AntiParallel3Constraint) or
              isinstance(source, Perp3Constraint)):
            points_solved.extend(self._solve_angle_constraint(source))
        elif isinstance(source, SymmetryConstraint):
            points_solved.extend(self.solve_symmetry_constraint(source))
        elif isinstance(source, ROCurvatureConstraint):
            points_solved.extend(self.solve_roc_constraint(source))

        # symmetry_constraints_to_re_solve = []
        # for roc_point in roc_points_solved:
        #     for geo_con in roc_point.geo_cons:
        #         if not isinstance(geo_con, SymmetryConstraint):
        #             continue
        #         if geo_con not in symmetry_constraints_to_re_solve:
        #             symmetry_constraints_to_re_solve.append(geo_con)
        #
        # symmetry_points_solved_second_pass = []
        # for symmetry_constraint in list(set(symmetry_constraints_to_re_solve)):
        #     symmetry_points_solved_second_pass.extend(self.solve_symmetry_constraint(symmetry_constraint))
        #
        # roc_constraints_to_re_solve = []
        # for symmetry_point in list(set(symmetry_points_solved_second_pass)):
        #     for geo_con in symmetry_point.geo_cons:
        #         if not isinstance(geo_con, ROCurvatureConstraint):
        #             continue
        #         if geo_con not in roc_constraints_to_re_solve:
        #             roc_constraints_to_re_solve.append(geo_con)
        #
        # roc_points_solved_second_pass = []
        # for roc_constraint in list(set(roc_constraints_to_re_solve)):
        #     roc_points_solved_second_pass.extend(self.solve_roc_constraint(roc_constraint))

        other_points_solved = self.solve_other_constraints(points_solved)
        other_points_solved = list(
            set(other_points_solved).union(
                set(symmetry_points_solved)).union(
                set(roc_points_solved))
        )
        # other_points_solved = list(
        #     set(other_points_solved).union(
        #         set(symmetry_points_solved)).union(
        #         set(roc_points_solved)).union(
        #         set(symmetry_points_solved_second_pass)).union(
        #         set(roc_points_solved_second_pass))
        # )

        points_solved = list(set(points_solved).union(set(other_points_solved)))

        return points_solved

    def translate_cluster(self, root: Point, dx: float, dy: float):
        if not root.root:
            raise ValueError("Cannot move a point that is not a root of a constraint cluster")
        points_solved = []
        for point in networkx.bfs_tree(self, source=root):
            point.x().set_value(point.x().value() + dx)
            point.y().set_value(point.y().value() + dy)
            if point not in points_solved:
                points_solved.append(point)
        self.solve_other_constraints(points_solved)
        return points_solved

    def rotate_cluster(self, rotation_handle: Point, new_rotation_handle_x: float = None,
                       new_rotation_handle_y: float = None,
                       new_rotation_angle: float = None):
        root = self._identify_root_from_rotation_handle(rotation_handle)
        if not root.root:
            raise ValueError("Cannot move a point that is not a root of a constraint cluster")
        old_rotation_handle_angle = root.measure_angle(rotation_handle)
        if new_rotation_angle is None:
            new_rotation_handle_angle = root.measure_angle(Point(new_rotation_handle_x, new_rotation_handle_y))
        else:
            new_rotation_handle_angle = new_rotation_angle
        delta_angle = new_rotation_handle_angle - old_rotation_handle_angle
        root_x = root.x().value()
        root_y = root.y().value()

        points_solved = []
        for point in networkx.bfs_tree(self, source=root):
            if point is root:
                continue
            old_x = point.x().value()
            old_y = point.y().value()
            new_x = (old_x - root_x) * np.cos(delta_angle) - (old_y - root_y) * np.sin(delta_angle) + root_x
            new_y = (old_x - root_x) * np.sin(delta_angle) + (old_y - root_y) * np.cos(delta_angle) + root_y
            point.x().set_value(new_x)
            point.y().set_value(new_y)
            if point not in points_solved:
                points_solved.append(point)
        self.solve_other_constraints(points_solved)
        return points_solved, root

    def solve_symmetry_constraint(self, constraint: SymmetryConstraint):
        # This code prevents recursion errors for the case where a new cluster is created from a symmetry constraint
        # target. The symmetry constraint only gets solved if it has not yet been added to the list of partial solves
        if constraint in self.partial_symmetry_solves:
            return []
        else:
            self.partial_symmetry_solves.append(constraint)

        x1, y1 = constraint.p1.x().value(), constraint.p1.y().value()
        x2, y2 = constraint.p2.x().value(), constraint.p2.y().value()
        x3, y3 = constraint.p3.x().value(), constraint.p3.y().value()
        line_angle = measure_abs_angle(x1, y1, x2, y2)
        tool_angle = measure_rel_angle3(x1, y1, x2, y2, x3, y3)
        if tool_angle < np.pi:
            mirror_angle = line_angle - np.pi / 2
        elif tool_angle > np.pi:
            mirror_angle = line_angle + np.pi / 2
        else:
            # Rare case where the point is coincident with the line: just make p4 = p3
            constraint.p4.request_move(constraint.p3.x().value(), constraint.p3.y().value(), force=True)
            return constraint.child_nodes

        mirror_distance = 2 * measure_point_line_distance_unsigned(x1, y1, x2, y2, x3, y3)
        constraint.p4.request_move(
            x3 + mirror_distance * np.cos(mirror_angle),
            y3 + mirror_distance * np.sin(mirror_angle), force=True
        )

        self.partial_symmetry_solves.remove(constraint)

        return constraint.child_nodes

    @staticmethod
    def solve_roc_constraint(constraint: ROCurvatureConstraint):

        def solve_for_single_curve(p_g1: Point, p_g2: Point, n: int):
            Lc = measure_curvature_length_bezier(
                constraint.curve_joint.x().value(), constraint.curve_joint.y().value(),
                p_g1.x().value(), p_g1.y().value(),
                p_g2.x().value(), p_g2.y().value(), constraint.param().value(), n
            )
            angle = p_g1.measure_angle(p_g2)
            p_g2.request_move(p_g1.x().value() + Lc * np.cos(angle),
                              p_g1.y().value() + Lc * np.sin(angle), force=True)

        def solve_for_single_curve_zero_curvature(p_g1: Point, p_g2: Point):
            p_g2.request_move(p_g1.x().value(), p_g1.y().value(), force=True)

        if constraint.curve_type_1 == "Bezier" and constraint.curve_type_2 == "LineSegment":
            solve_for_single_curve_zero_curvature(constraint.g1_point_curve_1, constraint.g2_point_curve_1)
            return constraint.child_nodes

        if constraint.curve_type_2 == "Bezier" and constraint.curve_type_1 == "LineSegment":
            solve_for_single_curve_zero_curvature(constraint.g1_point_curve_2, constraint.g2_point_curve_2)
            return constraint.child_nodes

        if constraint.curve_type_1 == "Bezier":
            if constraint.is_solving_allowed(constraint.g2_point_curve_1):
                solve_for_single_curve(constraint.g1_point_curve_1, constraint.g2_point_curve_1, constraint.curve_1.degree)
            else:
                R1 = ROCurvatureConstraint.calculate_curvature_data(constraint.curve_joint).R1
                constraint.param().set_value(R1, force=True)
        if constraint.curve_type_2 == "Bezier":
            if constraint.is_solving_allowed(constraint.g2_point_curve_2):
                solve_for_single_curve(constraint.g1_point_curve_2, constraint.g2_point_curve_2, constraint.curve_2.degree)
            else:
                R2 = ROCurvatureConstraint.calculate_curvature_data(constraint.curve_joint).R2
                constraint.param().set_value(R2, force=True)

        return constraint.child_nodes

    def solve_symmetry_constraints(self, points: typing.List[Point]):
        points_solved = []
        symmetry_constraints_solved = []
        for point in points:
            symmetry_constraints = [geo_con for geo_con in point.geo_cons if isinstance(geo_con, SymmetryConstraint)]
            for symmetry_constraint in symmetry_constraints:
                if symmetry_constraint in symmetry_constraints_solved:
                    continue
                symmetry_points_solved = self.solve_symmetry_constraint(symmetry_constraint)
                symmetry_constraints_solved.append(symmetry_constraint)
                for symmetry_point_solved in symmetry_points_solved:
                    if symmetry_point_solved in points_solved:
                        continue

                    points_solved.append(symmetry_point_solved)

        return points_solved

    def solve_roc_constraints(self, points: typing.List[Point]):
        points_solved = []
        roc_constraints_solved = []
        for point in points:
            roc_constraints = [geo_con for geo_con in point.geo_cons if isinstance(geo_con, ROCurvatureConstraint)]
            for roc_constraint in roc_constraints:
                if roc_constraint in roc_constraints_solved:
                    continue
                roc_points_solved = self.solve_roc_constraint(roc_constraint)
                roc_constraints_solved.append(roc_constraint)
                for roc_point_solved in roc_points_solved:
                    if roc_point_solved in points_solved:
                        continue

                    points_solved.append(roc_point_solved)

        return points_solved

    def solve_other_constraints(self, points: typing.List[Point]):
        roc_points_solved = self.solve_roc_constraints(points)
        symmetry_points_solved = self.solve_symmetry_constraints(points)
        return list(set(symmetry_points_solved).union(roc_points_solved))

    @staticmethod
    def update_canvas_items(points_solved: typing.List[Point]):

        curves_to_update = []
        for point in points_solved:
            if point.canvas_item is not None:
                point.canvas_item.updateCanvasItem(point.x().value(), point.y().value())

            for curve in point.curves:
                if curve not in curves_to_update:
                    curves_to_update.append(curve)

        airfoils_to_update = []
        for curve in curves_to_update:
            if curve.airfoil is not None and curve.airfoil not in airfoils_to_update:
                airfoils_to_update.append(curve.airfoil)
            curve.update()

        for airfoil in airfoils_to_update:
            airfoil.update_coords()
            if airfoil.canvas_item is not None:
                airfoil.canvas_item.generatePicture()

        constraints_to_update = []
        for point in points_solved:
            for geo_con in point.geo_cons:
                if geo_con not in constraints_to_update:
                    constraints_to_update.append(geo_con)

        for geo_con in constraints_to_update:
            if isinstance(geo_con, GeoCon) and geo_con.canvas_item is not None:
                geo_con.canvas_item.update()


class ConstraintValidationError(Exception):
    pass



from copy import deepcopy

from PyQt5.QtCore import Qt, QSize
from pyqtgraph import TableWidget

from pymead.core.geometry_collection import GeometryCollection


class BoundsValuesTable(TableWidget):
    def __init__(self, geo_col: GeometryCollection):
        super().__init__(editable=True, sortable=False)
        data = [[dv.name(), f"{dv.lower():.10f}", f"{dv.value():.10f}", f"{dv.upper():.10f}"]
                for dv in geo_col.container()["desvar"].values()]
        self.geo_col = geo_col
        self.initializing_data = True
        self.setData(data)
        self.setHorizontalHeaderLabels(["DesVar Name", "Lower Bound", "Value", "Upper Bound"])
        for column in (0, 2):
            self.makeColumnReadOnly(column)
        for column in range(self.columnCount()):
            self.setColumnWidth(column, 100)
        self.initializing_data = False

    def handleItemChanged(self, item):
        super().handleItemChanged(item)

        if not self.initializing_data:
            # Get the design variable from the geometry collection
            dv = self.geo_col.container()["desvar"][self.item(item.row(), 0).data(0)]

            # Check if the string input is castable to a float. If not, set the float_val to None
            if item.data(0).strip().replace(".", "").replace("-", "").isnumeric():
                float_val = float(item.data(0).strip())
            else:
                float_val = None

            # Lower bound
            if item.column() == 1:
                # Only change the design variable's lower bound if the input is castable to a float and is less than
                # or equal to the design variable's value
                if float_val is not None and float_val <= dv.value():
                    dv.set_lower(float_val)
                item.setData(0, f"{dv.lower():.10f}")

            # Upper bound
            elif item.column() == 3:
                # Only change the design variable's upper bound if the input is castable to a float and is greater than
                # or equal to the design variable's value
                if float_val is not None and float_val >= dv.value():
                    dv.set_upper(float_val)
                item.setData(0, f"{dv.upper():.10f}")

    def makeColumnReadOnly(self, column: int):
        for row in range(self.rowCount()):
            item = self.item(row, column)
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)

    def sizeHint(self):
        width = sum([self.columnWidth(i) for i in range(self.columnCount())])
        width += self.verticalHeader().sizeHint().width()
        width += self.verticalScrollBar().sizeHint().width()
        width += self.frameWidth() * 2
        return QSize(width + 12, self.height())

    def data(self):
        rows = list(range(self.rowCount()))
        columns = list(range(self.columnCount()))
        data = []
        if self.horizontalHeadersSet:
            row = []
            if self.verticalHeadersSet:
                row.append("")

            for c in columns:
                row.append(self.horizontalHeaderItem(c).text())
            data.append(row)

        for r in rows:
            row = []
            if self.verticalHeadersSet:
                row.append(self.verticalHeaderItem(r).text())
            for c in columns:
                item = self.item(r, c)
                if item is not None and item.value != "":
                    row.append(item.value)
                else:
                    row.append(None)
            data.append(row)

        return data



from pymoo.operators.sampling.lhs import LatinHypercubeSampling
from pymoo.core.problem import Problem
from pymead.utils.read_write_files import load_data, save_data
from pymead.core.mea import MEA
from matplotlib import pyplot as plt
from random import randint, random
import numpy as np
from copy import deepcopy
import os
import typing
from pymead.optimization.pop_chrom import Chromosome, Population
from abc import abstractmethod


class Sampling:
    def __init__(self, n_samples: int, norm_param_list: list):
        """
        High-level generic class for any sampling method.

        Parameters
        ==========
        n_samples: int
            Number of samples

        norm_param_list: list
            List of normalized parameter values (usually from ``extract_parameters`` in ``MEA``).
        """
        self.n_samples = n_samples
        self.norm_param_list = deepcopy(norm_param_list)

    @abstractmethod
    def sample(self) -> list or np.ndarray:
        pass


class ConstrictedRandomSampling(Sampling):
    def __init__(self, n_samples: int, norm_param_list: list, max_sampling_width: float):
        """
        A custom sampling method where the probability distribution for the new parameter value is uniform with a
        specified width on either side of the original parameter value. The probability of falling above or below the
        original parameter value is 0.5 and 0.5, respectively. The sampling width is truncated by the parameter bounds
        if necessary, retaining a probability of 0.5 on both sides of the original parameter value.

        Parameters
        ==========
        n_samples: int
            Number of samples

        norm_param_list: list
            List of normalized parameter values (usually from ``extract_parameters`` in ``MEA``).

        max_sampling_width: float
            Maximum distance of the new parameter value from the original parameter value. The actual new parameter
            value falls randomly in between the original parameter value and this maximum distance in either direction.
        """
        self.max_sampling_width = max_sampling_width
        super().__init__(n_samples, norm_param_list)

    def sample(self):
        """
        Randomly samples the design space.

        Returns
        =======
        list
            A new list of normalized parameter values to be used to update the airfoil system.
        """
        X_list = [self.norm_param_list]
        for i in range(self.n_samples - 1):
            individual = []
            norm_list_copy = deepcopy(self.norm_param_list)
            for p in norm_list_copy:
                sign_bool = randint(0, 1)
                perturbation = random()
                if sign_bool:
                    distance_to_upper_bound = 1 - p
                    if self.max_sampling_width < distance_to_upper_bound:
                        new_p = p + self.max_sampling_width * perturbation
                    else:
                        new_p = p + distance_to_upper_bound * perturbation
                else:
                    distance_to_lower_bound = p
                    if self.max_sampling_width < distance_to_lower_bound:
                        new_p = p - self.max_sampling_width * perturbation
                    else:
                        new_p = p - distance_to_lower_bound * perturbation
                individual.append(new_p)
            X_list.append(individual)
        return X_list


class PymooLHS(Sampling):
    def __init__(self, n_samples, norm_param_list):
        """
        Latin-Hypercube sampling method from the ``pymoo`` package.

        Parameters
        ==========
        n_samples: int
            Number of samples

        norm_param_list: list
            List of normalized parameter values (usually from ``extract_parameters`` in ``MEA``).
        """
        super().__init__(n_samples, norm_param_list)

    def sample(self):
        """
        Randomly samples the design space.

        Returns
        =======
        norm_param_list: np.ndarray
            1-D array of normalized parameter values (usually from ``extract_parameters`` in ``MEA``).
        """
        problem = Problem(n_var=len(self.norm_param_list), l=0.0, xu=1.0)
        sampling = LatinHypercubeSampling()
        return sampling._do(problem, self.n_samples)


def _run_analysis(analysis_dir: str, index: typing.Iterable, X_list, mea: dict, param_dict: dict, evaluate: bool = True,
                 save_coords: bool = False,
                 save_control_points: bool = False,
                 save_airfoil_state: bool = False):

    chromosomes = []

    if not os.path.exists(os.path.join(analysis_dir, 'analysis')):
        os.mkdir(os.path.join(analysis_dir, 'analysis'))

    for i in index:
        param_set = deepcopy(param_dict)
        param_set['mset_settings']['airfoil_analysis_dir'] = os.path.join(
            analysis_dir, 'analysis', f'analysis_{i}')
        param_set['mset_settings']['airfoil_coord_file_name'] = f'analysis_{i}'
        param_set['base_folder'] = os.path.join(analysis_dir, 'analysis')
        param_set['name'] = [f"analysis_{j}" for j in index]

        # parent_chromosomes.append(Chromosome(param_set=param_set, population_idx=s, mea=mea, X=X))
        chromosomes.append(Chromosome(param_dict=param_set, population_idx=i, mea=mea, genes=X_list[i],
                                      ga_settings=None, category=None, generation=0))

    population = Population(param_dict=param_dict, ga_settings=None, generation=0, parents=chromosomes,
                            mea=mea, verbose=True, skip_parent_assignment=False)
    population.generate_chromosomes_parallel()
    if save_coords:
        if not os.path.exists(os.path.join(analysis_dir, 'coords')):
            os.mkdir(os.path.join(analysis_dir, 'coords'))
        for idx, c in enumerate(population.population):
            save_data(c.coords, os.path.join(analysis_dir, 'coords', f'coords_{idx}.json'))
    if save_control_points:
        if not os.path.exists(os.path.join(analysis_dir, 'control_points')):
            os.mkdir(os.path.join(analysis_dir, 'control_points'))
        for idx, c in enumerate(population.population):
            save_data(c.control_points, os.path.join(analysis_dir,
                                                     'control_points', f'control_points_{idx}.json'))
    if save_airfoil_state:
        if not os.path.exists(os.path.join(analysis_dir, 'airfoil_state')):
            os.mkdir(os.path.join(analysis_dir, 'airfoil_state'))
        for idx, c in enumerate(population.population):
            save_data(c.airfoil_state, os.path.join(analysis_dir,
                                                    'airfoil_state', f'airfoil_state_{idx}.json'))

    if evaluate:
        population.eval_pop_fitness()
        # population_forces = {k: [c.forces[k] for c in population.population]
        #                      for k in population.population[0].forces.keys()}
        # print(f"{population_forces = }")
        # save_data(population_forces, post_process_force_file)


def _main():
    parametrization = r'C:\Users\mlauer2\Documents\pymead\pymead\pymead\tests\pai\root_underwing_opt\opt_runs\2023_03_16_A\pai_underwing.jmea'
    jmea_dict = load_data(parametrization)
    jmea_dict['airfoil_graphs_active'] = False
    analysis_dir = r'C:\Users\mlauer2\Documents\pymead\pymead\pymead\tests\pai\root_underwing_opt\opt_runs\test'
    param_dict = load_data(r'C:\Users\mlauer2\Documents\pymead\pymead\pymead\tests\pai\root_underwing_opt\opt_runs\test\param_dict.json')
    param_dict['mplot_settings']['flow_field'] = 2
    if param_dict['mses_settings']['multi_point_stencil'] is not None:
        for idx, stencil_var in enumerate(param_dict['mses_settings']['multi_point_stencil']):
            stencil_var['points'] = stencil_var['points'][:2]  # Stop at the design point (idx = 1)
            param_dict['mses_settings']['multi_point_stencil'][idx] = stencil_var
    param_dict['mses_settings']['timeout'] = 40.0
    mea = MEA.generate_from_param_dict(jmea_dict)
    norm_param_list, _ = mea.extract_parameters()
    sampling = ConstrictedRandomSampling(n_samples=50, norm_param_list=norm_param_list, max_sampling_width=0.08)
    fig, axs = plt.subplots()
    X_list = sampling.sample()
    for individual in X_list:
        mea.update_parameters(individual)
        color = np.random.choice(range(256), size=3) / 255
        for a in mea.airfoils.values():
            a.plot_airfoil(axs, color=color, lw=1.0)
    axs.set_aspect('equal')
    plt.show()
    # _run_analysis(analysis_dir=analysis_dir, index=[i for i in range(n_samples)], X_list=X_list,
    #              mea=jmea_dict, param_dict=param_dict)


if __name__ == '__main__':
    _main()



from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QTextEdit, QTextBrowser
from PyQt5.QtGui import QTextCursor, QColor, QFontDatabase, QDesktopServices


class ConsoleTextArea(QTextBrowser):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setReadOnly(True)
        self.setLineWrapMode(QTextEdit.NoWrap)
        # font = self.font()
        # print(QFontDatabase().families())
        # font.setFamily("DejaVu Sans Mono")
        # # font.setFamily("Courier")
        # font.setPointSize(6)
        # self
        # self.setFont(font)
        # self.setStyleSheet("""font: 10pt DejaVu Sans Mono;""")
        # prepend_html = f"<head><style>p {{font-family: DejaVu Sans Mono; font-size: 10pt}}</style>" \
        #                f"</head><body><p>&#8203;</p></body>"
        self.moveCursor(QTextCursor.End)
        # self.setTextColor(QColor("#13294B"))
        # self.setFontPointSize(5)
        self.setMinimumHeight(50)
        self.setOpenLinks(False)
        self.setOpenExternalLinks(True)

        self.document().setDefaultStyleSheet("""p, a {font-family: 'DejaVu Sans Mono'; font-size: 10pt}
                                                        a:link {color: orange;} 
                                                        a:active {color: green;} 
                                                        a:visited {color: hotpink;}
                                                        a:hover: {color: blue;}
                                                     """)

        def handle_links(url):
            if not url.scheme():
                url = QUrl.fromLocalFile(url.toString())
            QDesktopServices.openUrl(url)

        self.anchorClicked.connect(handle_links)



import numpy as np


def repair_example_sd7062(xy: np.ndarray):
    r"""
    ### Description:

    An example of how to repair a poorly-defined set of discrete airfoil coordinates.

    ### Args:

    `xy`: `np.ndarray` of `shape=(N, 2)` describing the set of "discrete" airfoil coordinates, where `N` is the number
    of coordinates, and the columns represent \(x\) and \(y\)

    ### Returns:

    The modified xy array
    """
    xy[-1, 0] = 1
    xy[-1, 1] = 0
    return xy



import logging
import multiprocessing.connection
import typing

import numpy as np

from copy import deepcopy

import os
import random

from pymead.core.geometry_collection import GeometryCollection
from pymead.optimization.opt_setup import CustomDisplay, TPAIOPT
from pymead.utils.read_write_files import load_data, save_data
from pymead.optimization.pop_chrom import Chromosome, Population
from pymead.optimization.objectives_and_constraints import Objective, Constraint
from pymead.optimization.sampling import ConstrictedRandomSampling
from pymead.optimization.opt_setup import termination_condition

import pymoo.core.population
from pymoo.algorithms.moo.unsga3 import UNSGA3
from pymoo.operators.mutation.pm import PolynomialMutation
from pymoo.operators.crossover.sbx import SimulatedBinaryCrossover
from pymoo.config import Config
from pymoo.core.evaluator import Evaluator
from pymoo.factory import get_reference_directions, get_decomposition
from pymoo.core.evaluator import set_cv


def shape_optimization(conn: multiprocessing.connection.Connection or None, param_dict: dict, opt_settings: dict,
                       geo_col_dict: dict,
                       objectives: typing.List[str], constraints: typing.List[str]):

    objectives = [Objective(func_str=func_str) for func_str in objectives]
    constraints = [Constraint(func_str=func_str) for func_str in constraints]

    # print(f"Shape optimization has PID {os.getpid()}")

    def start_message(warm_start: bool):
        first_word = "Resuming" if warm_start else "Beginning"
        return f"\n{first_word} aerodynamic shape optimization with {param_dict['num_processors']} processors..."

    def write_force_dict_to_file(forces_dict, file_name: str):
        forces_temp = deepcopy(forces_dict)
        if "Cp" in forces_temp.keys():
            for el in forces_temp["Cp"]:
                if isinstance(el, list):
                    for e in el:
                        for k_, v_ in e.items():
                            if isinstance(v_, np.ndarray):
                                e[k_] = v_.tolist()
                else:
                    for k_, v_ in el.items():
                        if isinstance(v_, np.ndarray):
                            el[k_] = v_.tolist()
        save_data(forces_temp, file_name)

    def read_force_dict_from_file(file_name: str):
        return load_data(file_name)

    def send_over_pipe(data: object):
        try:
            if conn is not None:
                conn.send(data)
        except BrokenPipeError:
            pass

    forces_dict = {} if not opt_settings["General Settings"]["warm_start_active"] \
        else read_force_dict_from_file(os.path.join(param_dict["opt_dir"], "force_history.json"))

    send_over_pipe(("text", start_message(opt_settings["General Settings"]["warm_start_active"])))

    Config.show_compile_hint = False
    forces = []
    ref_dirs = get_reference_directions("energy", param_dict['n_obj'], param_dict['n_ref_dirs'],
                                        seed=param_dict['seed'])
    geo_col = GeometryCollection.set_from_dict_rep(deepcopy(geo_col_dict))
    airfoil_name, mea_name = None, None
    if param_dict["tool"] == "XFOIL":
        airfoil_name = opt_settings["XFOIL"]["airfoil"]
    elif param_dict["tool"] == "MSES":
        mea_name = opt_settings["MSET"]["mea"]
    parameter_list = geo_col.extract_design_variable_values(bounds_normalized=True)
    num_parameters = len(parameter_list)

    send_over_pipe(("text", f"Number of design variables: {num_parameters}"))

    problem = TPAIOPT(n_var=param_dict['n_var'], n_obj=param_dict['n_obj'], n_constr=param_dict['n_constr'],
                      xl=param_dict['xl'], xu=param_dict['xu'], param_dict=param_dict)

    if not opt_settings['General Settings']['warm_start_active']:
        if param_dict['seed'] is not None:
            np.random.seed(param_dict['seed'])
            random.seed(param_dict['seed'])

        sampling = ConstrictedRandomSampling(n_samples=param_dict['n_offsprings'], norm_param_list=parameter_list,
                                             max_sampling_width=param_dict['max_sampling_width'])
        X_list = sampling.sample()
        parents = [Chromosome(geo_col_dict=geo_col_dict, param_dict=param_dict, generation=0,
                              airfoil_name=airfoil_name, mea_name=mea_name, population_idx=idx, genes=individual)
                   for idx, individual in enumerate(X_list)]
        population = Population(param_dict=param_dict, generation=0, parents=parents,
                                verbose=param_dict['verbose'])

        n_eval = population.eval_pop_fitness(sig=conn)
        print(f"Finished evaluating population fitness. Continuing...")

        new_X = None
        J = None
        G = None

        for chromosome in population.converged_chromosomes:
            forces.append(chromosome.forces)
            if new_X is None:
                if param_dict['n_offsprings'] > 1:
                    new_X = np.array([chromosome.genes])
                else:
                    new_X = np.array(chromosome.genes)
            else:
                new_X = np.row_stack((new_X, np.array(chromosome.genes)))
            # print(f"{self.objectives = }")
            # print(f"Before objective and constraint update, {chromosome.forces = }")
            for objective in objectives:
                objective.update(chromosome.forces)
            for constraint in constraints:
                constraint.update(chromosome.forces)
                print(f"{constraint.value = }")
            if J is None:
                J = np.array([obj.value for obj in objectives])
            else:
                J = np.row_stack((J, np.array([obj.value for obj in objectives])))
            if len(constraints) > 0:
                if G is None:
                    G = np.array([constraint.value for constraint in constraints])
                else:
                    G = np.row_stack((G, np.array([
                        constraint.value for constraint in constraints])))

            # print(f"{J = }, {self.objectives = }")

        if new_X is None:
            send_over_pipe(("text", f"Could not converge any individuals in the population. Optimization terminated."))
            return

        if new_X.ndim == 1:
            new_X = np.array([new_X])

        if J.ndim == 1:
            J = np.array([J])

        if len(constraints) > 0 and G.ndim == 1:
            G = np.array([G])

        pop_initial = pymoo.core.population.Population.new("X", new_X)
        # objectives
        pop_initial.set("F", J)
        # print(f"Initially, {J = }")
        if len(constraints) > 0:
            if G is not None:
                pop_initial.set("G", G)
        set_cv(pop_initial)
        for individual in pop_initial:
            individual.evaluated = {"F", "G", "CV", "feasible"}
        Evaluator(skip_already_evaluated=True).eval(problem, pop_initial)

        algorithm = UNSGA3(ref_dirs=ref_dirs, sampling=pop_initial,
                           n_offsprings=param_dict['n_offsprings'],
                           crossover=SimulatedBinaryCrossover(eta=param_dict['eta_crossover']),
                           mutation=PolynomialMutation(eta=param_dict['eta_mutation']))

        termination = termination_condition(param_dict)

        display = CustomDisplay()

        # prepare the algorithm to solve the specific problem (same arguments as for the minimize function)
        algorithm.setup(problem, termination, display=display, seed=param_dict['seed'], verbose=True,
                        save_history=False)

        algorithm.evaluator.n_eval += n_eval

        # save_data(algorithm, os.path.join(param_dict['opt_dir'], 'algorithm_gen_0.pkl'))

        # np.save('checkpoint', algorithm)
        # until the algorithm has no terminated
        n_generation = 0
    else:
        logging.debug('Starting from where we left off...')
        warm_start_index = param_dict['warm_start_generation']
        n_generation = warm_start_index
        warm_start_alg_file = os.path.join(opt_settings['General Settings']['warm_start_dir'],
                                           f'algorithm_gen_{warm_start_index}.pkl')
        algorithm = load_data(warm_start_alg_file)
        read_force_dict_from_file(os.path.join(param_dict["opt_dir"], "force_history.json"))
        logging.debug(f'Loaded {warm_start_alg_file}.')
        if not opt_settings['General Settings']['use_initial_settings']:
            # Currently only set up to change n_offsprings
            previous_offsprings = deepcopy(algorithm.n_offsprings)
            algorithm.n_offsprings = opt_settings['Genetic Algorithm']['n_offspring']
            algorithm.problem.param_dict['n_offsprings'] = algorithm.n_offsprings
            if previous_offsprings != algorithm.n_offsprings:
                logging.debug(f'Number of offspring changed from {previous_offsprings} '
                              f'to {algorithm.n_offsprings}.')
        term = deepcopy(algorithm.termination.terminations)
        term = list(term)
        term[0].n_max_gen = param_dict['n_max_gen']
        term = tuple(term)
        algorithm.termination.terminations = term
        algorithm.has_terminated = False

    while algorithm.has_next():

        logging.debug(f'Asking the algorithm to get the next population...')

        pop = algorithm.ask()

        logging.debug(f'Acquired new population. pop = {pop}')

        n_generation += 1

        if n_generation > 1:

            logging.debug(f'Starting generation {n_generation}...')

            forces = []

            # evaluate (objective function value arrays must be numpy column vectors)
            X = pop.get("X")
            logging.debug(f'Input matrix has shape {X.shape} ({X.shape[0]} chromosomes with {X.shape[1]} genes).')
            new_X = None
            J = None
            G = None

            parents = [Chromosome(param_dict=param_dict, generation=n_generation, population_idx=idx,
                                  airfoil_name=airfoil_name, mea_name=mea_name,
                                  geo_col_dict=geo_col_dict, genes=individual) for idx, individual in enumerate(X)]
            population = Population(problem.param_dict, generation=n_generation,
                                    parents=parents, verbose=param_dict['verbose'])
            n_eval = population.eval_pop_fitness(sig=conn)

            for chromosome in population.converged_chromosomes:
                forces.append(chromosome.forces)
                if new_X is None:
                    if param_dict['n_offsprings'] > 1:
                        new_X = np.array([chromosome.genes])
                    else:
                        new_X = np.array(chromosome.genes)
                else:
                    new_X = np.row_stack((new_X, np.array(chromosome.genes)))
                for objective in objectives:
                    objective.update(chromosome.forces)
                for constraint in constraints:
                    constraint.update(chromosome.forces)
                if J is None:
                    J = np.array([obj.value for obj in objectives])
                else:
                    J = np.row_stack((J, np.array([obj.value for obj in objectives])))
                if len(constraints) > 0:
                    if G is None:
                        G = np.array([constraint.value for constraint in constraints])
                    else:
                        G = np.row_stack((G, np.array([
                            constraint.value for constraint in constraints])))

                # print(f"{J = }, {self.objectives = }")

            algorithm.evaluator.n_eval += n_eval

            if new_X is None:
                send_over_pipe(
                    ("text", f"Could not converge any individuals in the population. Optimization terminated."))
                return

            for idx in range(param_dict['n_offsprings'] - len(new_X)):
                # f1 = np.append(f1, np.array([1000.0]))
                # f2 = np.append(f2, np.array([1000.0]))
                new_X = np.row_stack((new_X, 9999 * np.ones(param_dict['n_var'])))
                J = np.row_stack((J, 1000.0 * np.ones(param_dict['n_obj'])))
                if len(constraints) > 0:
                    G = np.row_stack((G, 1000.0 * np.ones(param_dict['n_constr'])))

            if new_X.ndim == 1:
                new_X = np.array([new_X])

            if J.ndim == 1:
                J = np.array([J])

            if len(constraints) > 0 and G.ndim == 1:
                G = np.array([G])

            pop.set("X", new_X)

            # objectives
            pop.set("F", J)

            # print(f"{pop.get('F') = }")

            # for constraints
            if len(constraints) > 0:
                pop.set("G", G)

            # this line is necessary to set the CV and feasbility status - even for unconstrained
            set_cv(pop)

        # returned the evaluated individuals which have been evaluated or even modified
        # print(f"{pop.get('X') = }, {pop.get('F') = }")
        algorithm.tell(infills=pop)

        # print(f"{algorithm.opt.get('F') = }")

        warm_start_gen = None
        if opt_settings["General Settings"]["warm_start_active"]:
            warm_start_gen = param_dict["warm_start_generation"]

        send_over_pipe(("opt_progress", {"text": algorithm.display.progress_dict, "completed": not algorithm.has_next(),
                                         "warm_start_gen": warm_start_gen}))

        # print(f"{algorithm.display.progress_dict = }")

        if len(objectives) == 1:
            if n_generation > 1:
                X = algorithm.opt.get("X")[0]
            else:
                X = algorithm.pop.get("X")[0, :]
        else:
            if n_generation > 1:
                X = algorithm.opt.get("X")
                F = algorithm.opt.get("F")
            else:
                X = algorithm.pop.get("X")
                F = algorithm.pop.get("F")

            # Use a "50-50" decomposition to get a representative optimal solution for plotting
            weights = np.array([1.0 / len(objectives) for _ in objectives])
            decomp = get_decomposition("asf")
            I = decomp.do(F, weights).argmin()
            X = X[I, :]

        best_in_previous_generation = False
        forces_index = 0
        try:
            forces_index = np.where((new_X == X).all(axis=1))[0][0]
        except IndexError:
            best_in_previous_generation = True

        if best_in_previous_generation:
            for k, v in forces_dict.items():
                if k not in forces_dict.keys():
                    forces_dict[k] = []
                forces_dict[k].append(v[-1])
        else:
            best_forces = forces[forces_index]
            for k, v in best_forces.items():
                if param_dict['tool'] in ['xfoil', 'XFOIL', 'mses', 'MSES', 'Mses']:
                    if k not in ['converged', 'timed_out', 'errored_out']:
                        if k not in forces_dict.keys():
                            forces_dict[k] = []
                        forces_dict[k].append(v)

        write_force_dict_to_file(forces_dict, os.path.join(param_dict["opt_dir"], "force_history.json"))

        geo_col.assign_design_variable_values(X.tolist(), bounds_normalized=True)
        mea = None if mea_name is None else geo_col.container()["mea"][mea_name]
        airfoil = None if airfoil_name is None else geo_col.container()["airfoils"][airfoil_name]
        if airfoil is not None:
            coords = [airfoil.get_coords_selig_format()]
        else:
            coords = mea.get_coords_list()

        send_over_pipe(("airfoil_coords", coords))

        norm_val_list = geo_col.extract_design_variable_values(bounds_normalized=True)

        send_over_pipe(("parallel_coords", (norm_val_list, [desvar for desvar in geo_col.container()["desvar"]])))

        if param_dict["tool"] == "XFOIL":
            Cp = forces_dict["Cp"][-1]
            if isinstance(Cp, list):
                Cp = Cp[param_dict["design_idx"]]
            send_over_pipe(("cp_xfoil", Cp))

            Cd = forces_dict['Cd']
            Cdp = forces_dict['Cdp']
            Cdf = forces_dict['Cdf']
            Cd = Cd if not isinstance(Cd[0], list) else [d[param_dict["design_idx"]] for d in Cd]
            Cdp = Cdp if not isinstance(Cdp[0], list) else [d[param_dict["design_idx"]] for d in Cdp]
            Cdf = Cdf if not isinstance(Cdf[0], list) else [d[param_dict["design_idx"]] for d in Cdf]
            send_over_pipe(("drag_xfoil", (Cd, Cdp, Cdf)))
        elif param_dict["tool"] == "MSES":
            Cp = forces_dict['BL'][-1] if isinstance(
                forces_dict['BL'][-1][0], dict) else forces_dict['BL'][-1][param_dict["design_idx"]]
            send_over_pipe(("cp_mses", Cp))

            Cd = forces_dict['Cd']
            Cdp = forces_dict['Cdp']
            Cdf = forces_dict['Cdf']
            Cdv = forces_dict['Cdv']
            Cdw = forces_dict['Cdw']
            Cd = Cd if not isinstance(Cd[0], list) else [d[param_dict["design_idx"]] for d in Cd]
            Cdp = Cdp if not isinstance(Cdp[0], list) else [d[param_dict["design_idx"]] for d in Cdp]
            Cdf = Cdf if not isinstance(Cdf[0], list) else [d[param_dict["design_idx"]] for d in Cdf]
            Cdv = Cdv if not isinstance(Cdv[0], list) else [d[param_dict["design_idx"]] for d in Cdv]
            Cdw = Cdw if not isinstance(Cdw[0], list) else [d[param_dict["design_idx"]] for d in Cdw]
            send_over_pipe(("drag_mses", (Cd, Cdp, Cdf, Cdv, Cdw)))

        if n_generation % param_dict['algorithm_save_frequency'] == 0:
            save_data(algorithm, os.path.join(param_dict['opt_dir'], f'algorithm_gen_{n_generation}.pkl'))

    # obtain the result objective from the algorithm
    res = algorithm.result()
    save_data(res, os.path.join(param_dict['opt_dir'], 'res.pkl'))
    write_force_dict_to_file(forces_dict, os.path.join(param_dict["opt_dir"], "force_history.json"))
    np.savetxt(os.path.join(param_dict['opt_dir'], 'opt_X.dat'), res.X)

    send_over_pipe(("finished", None))
    # self.save_opt_plots(param_dict['opt_dir'])  # not working at the moment



import sys
import typing

import networkx
import numpy as np

from pymead.core.param import LengthParam
from pymead.core.pymead_obj import PymeadObj


class Point(PymeadObj):
    """
    The ``Point`` is the lowest-level geometry object in `pymead`. All curves in `pymead` are tied directly to ``Point``
    objects. For example, instances of this class are used to define the endpoints of finite lines and control points of
    Bézier curves.
    """
    def __init__(self, x: float, y: float, name: str or None = None, setting_from_geo_col: bool = False):
        super().__init__(sub_container="points")
        self._x = None
        self._y = None
        self.gcs = None
        self.root = False
        self.rotation_handle = False
        self.rotation_param = None
        self.geo_cons = []
        self.dims = []
        self.curves = []
        self.setting_from_geo_col = setting_from_geo_col
        name = "Point-1" if name is None else name
        self.set_name(name)
        self.set_x(x)
        self.set_y(y)

    def x(self):
        """
        Getter for the point's ``x`` parameter.

        Returns
        -------
        LengthParam
            The point's ``x`` parameter
        """
        return self._x

    def y(self):
        """
        Getter for the point's ``y`` parameter.

        Returns
        -------
        LengthParam
            The point's ``y`` parameter
        """
        return self._y

    def set_x(self, x: LengthParam or float):
        self._x = x if isinstance(x, LengthParam) else LengthParam(
            value=x, name=self.name() + ".x", setting_from_geo_col=self.setting_from_geo_col, point=self)
        if self not in self._x.geo_objs:
            self._x.geo_objs.append(self)
        self._x.point = self

    def set_y(self, y: LengthParam or float):
        self._y = y if isinstance(y, LengthParam) else LengthParam(
            value=y, name=self.name() + ".y", setting_from_geo_col=self.setting_from_geo_col, point=self)
        if self not in self._y.geo_objs:
            self._y.geo_objs.append(self)
        self._y.point = self

    def set_name(self, name: str):
        """
        Extends the base ``set_name()`` by also renaming the point's ``x`` and ``y`` parameters.

        Parameters
        ----------
        name: str
            Name for the point
        """
        # Rename the x and y parameters of the Point
        if self.x() is not None:
            self.x().set_name(f"{name}.x")
        if self.y() is not None:
            self.y().set_name(f"{name}.y")

        super().set_name(name)

    def as_array(self):
        """
        Gives a one-dimensional, two-element array representation of the point (:math:`x` and :math:`y` values)

        Returns
        -------
        np.ndarray
            One-dimensional array containing the point's :math:`x` and :math:`y` values
        """
        return np.array([self.x().value(), self.y().value()])

    @classmethod
    def generate_from_array(cls, arr: np.ndarray, name: str or None = None):
        if arr.ndim != 1:
            raise ValueError("Points can only be generated from 1-dimensional arrays")
        return cls(x=arr[0], y=arr[1], name=name)

    def measure_distance(self, other: "Point") -> float:
        """
        Measures the distance from this point to another point.

        Parameters
        ----------
        other: Point
            Other point (the endpoint of the line whose distance is measured)

        Returns
        -------
        float
            The distance between ``self`` and ``other``
        """
        # Note: the quotes around "Point" are necessary here because this type hint is a forward reference.
        # This means that an object of type <Class> is specified as a hint somewhere inside the definition for <Class>
        return np.hypot(other.x().value() - self.x().value(), other.y().value() - self.y().value())

    def measure_angle(self, other: "Point"):
        """
        Measures the angle (in radians) of the line starting at this point and ending at ``other``

        Parameters
        ----------
        other: Point
            Other point (the endpoint of the line whose angle is measured)

        Returns
        -------
        float
            The angle of the line connecting ``self`` and ``other``
        """
        return np.arctan2(other.y().value() - self.y().value(), other.x().value() - self.x().value())

    def is_coincident(self, other: "Point", rtol: float = 1.0e-14) -> bool:
        """
        Determines whether this point is coincident with another point (within a tight tolerance)

        Parameters
        ----------
        other: Point
            Other point to test for coincidence with this point

        rtol: float
            Relative tolerance used for the coincidence test. Default: ``1e-14``

        Returns
        -------
        bool
            Whether the points are coincident
        """
        dist = self.measure_distance(other)
        if np.isclose(dist, 0.0, rtol=rtol):
            return True
        return False

    def _is_symmetry_123_and_no_edges(self) -> list or bool:
        """
        Checks if this point is a member of a symmetry constraint (but not the target of the symmetry constraint) and
        has no attached edges in the constraint graph. Used to determine if the point is allowed to move in
        ``request_move()``.

        Returns
        -------
        list or bool
            The list of symmetry constraints this point is a member of if the above conditions are met,
            ``False`` otherwise
        """
        if self.gcs is None:
            return False
        if len([edge for edge in self.gcs.in_edges(nbunch=self)]) != 0:
            return False
        if len([edge for edge in self.gcs.out_edges(nbunch=self)]) != 0:
            return False
        symmetry_constraints = []
        for geo_con in self.geo_cons:
            if geo_con.__class__.__name__ == "SymmetryConstraint":
                symmetry_constraints.append(geo_con)
        for symmetry_constraint in symmetry_constraints:
            if self is symmetry_constraint.p4:
                return False
        return symmetry_constraints

    def _is_symmetry_target_and_has_edges(self) -> bool:
        if self.gcs is None:
            return False
        if (not [edge for edge in self.gcs.in_edges(nbunch=self)] and
            not [edge for edge in self.gcs.out_edges(nbunch=self)]):
            return False
        symmetry_constraints = []
        for geo_con in self.geo_cons:
            if geo_con.__class__.__name__ == "SymmetryConstraint":
                symmetry_constraints.append(geo_con)
        if not symmetry_constraints:
            return False
        if not any([self is symmetry_constraint.p4 for symmetry_constraint in symmetry_constraints]):
            return False
        return True

    def is_movement_allowed(self) -> bool:
        """
        This method determines if movement is allowed for the point. Movement is allowed in these cases:

        - Where the constraint solver has not been set or there are no geometric constraints attached
        - Where the point is a root or rotation handle of a constraint cluster. If a rotation handle, movement is
          allowed, but the movement gets accepted as a rotation about the root point with a fixed distance to the
          root point
        - Where the point is one of the first three out of the four points in a symmetry constraint, and no edges are
          attached to this point in the constraint graph

        Returns
        -------
        bool
            ``True`` if movement is allowed for this point, ``False`` otherwise
        """
        if any([curve.__class__.__name__ == "PolyLine" for curve in self.curves]):
            return False
        if self.gcs is None:
            return True
        if self.gcs is not None and len(self.geo_cons) == 0:
            return True
        if self.gcs is not None and self.root:
            return True
        if self.gcs is not None and self.rotation_handle:
            # In this case, movement is allowed, but movements get translated to a rotation about the root point with
            # a fixed distance to the root point
            return True
        if self._is_symmetry_123_and_no_edges():
            return True
        return False

    def request_move(self, xp: float, yp: float, force: bool = False):
        """
        Updates the location of the point and updates any curves and canvas items associated with the point movement.

        Parameters
        ----------
        xp: float
            New :math:`x`-value for the point

        yp: float
            New :math:`y`-value for the point

        force: bool
            Force the movement of this point. Overrides ``is_movement_allowed``.

        Warning
        -------
        The ``force`` keyword argument should **never** be called directly from the API, or unexpected behavior
        may result. This argument is used in the backend code for the constraint solver in the symmetry and curvature
        constraints.
        """

        if not self.is_movement_allowed() and not force:
            return

        points_to_update = None
        if self.root:
            points_to_update = self.gcs.translate_cluster(self, dx=xp - self.x().value(), dy=yp - self.y().value())
            constraints_to_update = []
            for point in networkx.dfs_preorder_nodes(self.gcs, source=self):
                for geo_con in point.geo_cons:
                    if geo_con not in constraints_to_update:
                        constraints_to_update.append(geo_con)

            for geo_con in constraints_to_update:
                if geo_con.canvas_item is not None:
                    geo_con.canvas_item.update()
        elif self.rotation_handle:
            if self.rotation_param is None:
                return
            self.rotation_param.set_value(self.rotation_param.root.measure_angle(Point(xp, yp)))
        else:
            self.x().set_value(xp)
            self.y().set_value(yp)
            points_to_update = [self]
            symmetry_constraints = self._is_symmetry_123_and_no_edges()
            if symmetry_constraints:
                for symmetry_constraint in symmetry_constraints:
                    self.gcs.solve_symmetry_constraint(symmetry_constraint)
                    points_to_update.extend(symmetry_constraint.child_nodes)
                points_to_update = list(set(points_to_update))  # Get only the unique points
                for symmetry_constraint in symmetry_constraints:
                    if symmetry_constraint.canvas_item is not None:
                        symmetry_constraint.canvas_item.update()

        if self._is_symmetry_target_and_has_edges():
            points_solved = []
            for gc in self.geo_cons:
                points_solved.extend(self.gcs.solve(gc))
            self.gcs.update_canvas_items(list(set(points_solved)))

        # Update the GUI object, if there is one
        if self.canvas_item is not None:
            self.canvas_item.updateCanvasItem(self.x().value(), self.y().value())

        if points_to_update is None:
            return

        curves_to_update = []
        for point in points_to_update:
            if point.canvas_item is not None:
                point.canvas_item.updateCanvasItem(point.x().value(), point.y().value())

            for curve in point.curves:
                if curve not in curves_to_update:
                    curves_to_update.append(curve)

        airfoils_to_update = []
        for curve in curves_to_update:
            if curve.airfoil is not None and curve.airfoil not in airfoils_to_update:
                airfoils_to_update.append(curve.airfoil)
            curve.update()

        for airfoil in airfoils_to_update:
            airfoil.update_coords()
            if airfoil.canvas_item is not None:
                airfoil.canvas_item.generatePicture()

    def __repr__(self):
        return f"Point {self.name()}<x={self.x().value():.6f}, y={self.y().value():.6f}>"

    def get_dict_rep(self):
        return {"x": float(self.x().value()), "y": float(self.y().value())}


class PointSequence:
    def __init__(self, points: typing.List[Point]):
        self._points = None
        self.set_points(points)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return self.generate_from_slice(self, idx)
        else:
            return self.points()[idx]

    def __setitem__(self, idx, val):
        self.points()[idx] = val

    def __len__(self):
        return len(self.points())

    @classmethod
    def generate_from_slice(cls, original_point_seq, s):
        return cls(points=original_point_seq.points()[s].copy())

    @classmethod
    def generate_from_array(cls, arr: np.ndarray):
        if arr.shape[1] != 2:
            raise ValueError(f"Array must have two columns, x and y. Found {arr.shape[1]} columns.")
        return cls(points=[Point(x=x, y=y, name=f"PointFromArray-Index{idx}")
                           for idx, (x, y) in enumerate(zip(arr[:, 0], arr[:, 1]))])

    def points(self):
        return self._points

    def point_idx_from_ref(self, point: Point):
        return self.points().index(point)

    def reverse(self):
        self.points().reverse()

    def set_points(self, points: typing.List[Point]):
        self._points = points

    def insert_point(self, idx: int, point: Point):
        self._points.insert(idx, point)

    def append_point(self, point: Point):
        self._points.append(point)

    def remove_point(self, idx: int):
        self._points.pop(idx)

    def as_array(self):
        return np.array([[p.x().value(), p.y().value()] for p in self.points()])

    def extract_subsequence(self, indices: list):
        return PointSequence(points=[self.points()[idx] for idx in indices])



import numpy as np
import pyqtgraph as pg
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QWidget, QGridLayout

from pymead.core import UNITS

from pymead.utils.misc import get_setting


class AnalysisGraph:
    def __init__(self, theme: dict, pen=None, size: tuple = (1000, 300), background_color: str = 'w'):
        pg.setConfigOptions(antialias=True)

        if pen is None:
            pen = pg.mkPen(color='cornflowerblue', width=2)

        self.w = pg.GraphicsLayoutWidget(show=True, size=size)

        self.v = self.w.addPlot(pen=pen)
        self.v.invertY(True)
        self.legend = self.v.addLegend(offset=(300, 20))
        self.set_formatting(theme=theme)

    def set_background(self, background_color: str):
        self.w.setBackground(background_color)

    def set_formatting(self, theme: dict):
        self.w.setBackground(theme["graph-background-color"])
        label_font = f"{get_setting('axis-label-point-size')}pt {get_setting('axis-label-font-family')}"
        self.v.setLabel(axis="bottom", text=f"x [{UNITS.current_length_unit()}]", font=label_font,
                           color=theme["main-color"])
        self.v.setLabel(axis="left", text=f"Pressure Coefficient", font=label_font, color=theme["main-color"])
        tick_font = QFont(get_setting("axis-tick-font-family"), get_setting("axis-tick-point-size"))
        self.v.getAxis("bottom").setTickFont(tick_font)
        self.v.getAxis("left").setTickFont(tick_font)
        self.v.getAxis("bottom").setTextPen(theme["main-color"])
        self.v.getAxis("left").setTextPen(theme["main-color"])

    def set_legend_label_format(self, theme: dict):
        for _, label in self.legend.items:
            label.item.setHtml(f"<span style='font-family: DejaVu Sans; color: "
                               f"{theme['main-color']}; size: 8pt'>{label.text}</span>")


class ResidualGraph:
    def __init__(self, theme: dict, pen=None, size: tuple = (1000, 300)):
        pg.setConfigOptions(antialias=True)

        if pen is None:
            pen = pg.mkPen(color='cornflowerblue', width=2)

        self.w = pg.GraphicsLayoutWidget(show=True, size=size)

        self.v = self.w.addPlot(pen=pen)
        self.v.setLogMode(x=False, y=True)
        self.legend = self.v.addLegend(offset=(-5, 5))
        target_pen = self.make_target_pen(theme)

        # Add a line at y=1.0e-5 representing the target value for the residuals
        self.target_line = self.v.addLine(y=-5, pen=target_pen)

        # Dummy item to get the target line to add to the legend
        self.target_line_dummy_plot = self.v.plot(pen=target_pen, name="Target")

        # Set up the residual plot lines
        self.plot_items = [
            self.v.plot(pen=pg.mkPen(color=theme["residual-color-1"]), name="Density"),  # rms(dR): DRRMS in STATE.INC
            self.v.plot(pen=pg.mkPen(color=theme["residual-color-2"]), name="Grid Node"),  # rms(dA): DNRMS in STATE.INC
            self.v.plot(pen=pg.mkPen(color=theme["residual-color-3"]), name="Viscosity")  # rms(dV): DVRMS in STATE.INC
        ]

        # Set the formatting for the graph based on the current theme
        self.set_formatting(theme=theme)

    @staticmethod
    def make_target_pen(theme: dict):
        target_pen = pg.mkPen(color=theme["residual-target-color"])
        target_pen.setStyle(Qt.DashLine)
        return target_pen

    def set_background(self, background_color: str):
        self.w.setBackground(background_color)

    def set_formatting(self, theme: dict):
        self.w.setBackground(theme["graph-background-color"])
        label_font = f"{get_setting('axis-label-point-size')}pt {get_setting('axis-label-font-family')}"
        self.v.setLabel(axis="bottom", text=f"Iteration", font=label_font,
                           color=theme["main-color"])
        self.v.setLabel(axis="left", text=f"Residual Value", font=label_font, color=theme["main-color"])
        tick_font = QFont(get_setting("axis-tick-font-family"), get_setting("axis-tick-point-size"))
        self.v.getAxis("bottom").setTickFont(tick_font)
        self.v.getAxis("left").setTickFont(tick_font)
        self.v.getAxis("bottom").setTextPen(theme["main-color"])
        self.v.getAxis("left").setTextPen(theme["main-color"])
        for idx, plot_item in enumerate(self.plot_items):
            plot_item.setPen(pg.mkPen(color=theme[f"residual-color-{idx+1}"]))
        target_pen = self.make_target_pen(theme)
        self.target_line.setPen(target_pen)
        self.target_line_dummy_plot.setPen(target_pen)

    def set_legend_label_format(self, theme: dict):
        for _, label in self.legend.items:
            label.item.setHtml(f"<span style='font-family: DejaVu Sans; color: "
                               f"{theme['main-color']}; size: 8pt'>{label.text}</span>")


class SinglePolarGraph:
    def __init__(self, theme: dict, graph_color_key: str, x_axis_label: str, y_axis_label: str,
                 pen=None, size: tuple = (350, 300)):
        pg.setConfigOptions(antialias=True)

        if pen is None:
            pen = pg.mkPen(color='cornflowerblue', width=2)

        self.w = pg.GraphicsLayoutWidget(show=True, size=size)

        self.v = self.w.addPlot(pen=pen)

        # Set up the line
        self.plot_items = [self.v.plot(pen=pg.mkPen(color=theme[graph_color_key]))]

        self.x_axis_label = x_axis_label
        self.y_axis_label = y_axis_label
        self.graph_color_key = graph_color_key

        # Set the formatting for the graph based on the current theme
        self.set_formatting(theme=theme)

    def set_background(self, background_color: str):
        self.w.setBackground(background_color)

    def set_formatting(self, theme: dict):
        self.w.setBackground(theme["graph-background-color"])
        label_font = f"{get_setting('axis-label-point-size')}pt {get_setting('axis-label-font-family')}"
        self.v.setLabel(axis="bottom", text=self.x_axis_label, font=label_font,
                        color=theme["main-color"])
        self.v.setLabel(axis="left", text=self.y_axis_label, font=label_font, color=theme["main-color"])
        tick_font = QFont(get_setting("axis-tick-font-family"), get_setting("axis-tick-point-size"))
        self.v.getAxis("bottom").setTickFont(tick_font)
        self.v.getAxis("left").setTickFont(tick_font)
        self.v.getAxis("bottom").setTextPen(theme["main-color"])
        self.v.getAxis("left").setTextPen(theme["main-color"])
        for idx, plot_item in enumerate(self.plot_items):
            plot_item.setPen(pg.mkPen(color=theme[self.graph_color_key]))


class PolarGraphCollection(QWidget):
    def __init__(self, theme: dict):
        self.polar_graphs = [
            SinglePolarGraph(theme=theme, graph_color_key="polar-color-1",
                             x_axis_label="&alpha; (&deg;)", y_axis_label="C<sub>l</sub>"),
            SinglePolarGraph(theme=theme, graph_color_key="polar-color-2",
                             x_axis_label="C<sub>d</sub>", y_axis_label="C<sub>l</sub>"),
            SinglePolarGraph(theme=theme, graph_color_key="polar-color-3",
                             x_axis_label="&alpha; (&deg;)", y_axis_label="L/D"),
            SinglePolarGraph(theme=theme, graph_color_key="polar-color-4",
                             x_axis_label="&alpha; (&deg;)", y_axis_label="C<sub>m</sub>")
        ]
        super().__init__(parent=None)
        self.lay = QGridLayout()
        self.setLayout(self.lay)
        self.lay.addWidget(self.polar_graphs[0].w, 0, 0)
        self.lay.addWidget(self.polar_graphs[1].w, 0, 1)
        self.lay.addWidget(self.polar_graphs[2].w, 1, 0)
        self.lay.addWidget(self.polar_graphs[3].w, 1, 1)

    def set_background(self, background_color: str):
        for polar_graph in self.polar_graphs:
            polar_graph.set_background(background_color)

    def set_formatting(self, theme: dict):
        for polar_graph in self.polar_graphs:
            polar_graph.set_formatting(theme)

    def set_data(self, aero_data: dict):
        lift_to_drag_ratio = (np.array(aero_data["Cl"]) / np.array(aero_data["Cd"])).tolist()
        self.polar_graphs[0].plot_items[0].setData(x=aero_data["alf"], y=aero_data["Cl"])
        self.polar_graphs[1].plot_items[0].setData(x=aero_data["Cd"], y=aero_data["Cl"])
        self.polar_graphs[2].plot_items[0].setData(x=aero_data["alf"], y=lift_to_drag_ratio)
        self.polar_graphs[3].plot_items[0].setData(x=aero_data["alf"], y=aero_data["Cm"])

    def clear_data(self):
        for polar_graph in self.polar_graphs:
            polar_graph.plot_items[0].setData([], [])

    def toggle_grid(self):
        for polar_graph in self.polar_graphs:
            x_state = polar_graph.v.ctrl.xGridCheck.checkState()
            y_state = polar_graph.v.ctrl.yGridCheck.checkState()
            if x_state or y_state:
                polar_graph.v.showGrid(x=False, y=False)
            else:
                polar_graph.v.showGrid(x=True, y=True)



import math

from decimal import Decimal


def nchoosek(n: int, k: int):
    r"""
    Computes the mathematical combination

    .. math::

      n \choose k

    Parameters
    ==========
    n: int
      Number of elements in the set

    k: int
      Number of items to select from the set

    Returns
    =======
    :math:`n \choose k`
    """
    f = math.factorial
    return float(Decimal(f(n)) / Decimal(f(k)) / Decimal(f(n - k)))





def viscosity_calculator(T: float, rho: float or None = None, input_units: str = 'K') -> float:
    r"""
    Used to calculate the viscosity from temperature using Sutherland's law for dialogs in the GUI

    Parameters
    ==========
    T: float
        The input temperature (can be in units of :math:`K`, :math:`^{\circ}C`, or :math:`^{\circ}F`)

    rho: float or None
      The density, specified if an output of kinematic viscosity is desired instead of dynamic viscosity.
      Default: ``None``.

    input_units: str
      Input units. Must be one of ``"K"``, ``"C"``, or ``"F"``. Default: ``"K"``

    Returns
    =======
    float
        The dynamic viscosity if the density is not specified, or the kinematic viscosity if the density *is* specified
    """
    if input_units == 'K':
        pass
    elif input_units == 'C':
        T = T + 273.15  # convert from Celsius to Kelvin
    elif input_units == 'F':
        T = (T - 32) * 5 / 9 + 273.15  # convert from Fahrenheit to Kelvin
    else:
        raise ValueError('Invalid selection for \'input_units\' (must be one of \'K\', \'C\', or \'F\')')
    mu_ref = 1.716e-5
    T_ref = 273.15
    S = 110.4
    mu = mu_ref * (T / T_ref) ** (3 / 2) * (T_ref + S) / (T + S)
    if rho is None:
        return mu
    else:
        nu = mu / rho
        return nu



r"""
Welcome
=======
To the documentation page for pymead, an object-oriented Python 3 package for single- and multi-element
Bézier-parametrized airfoil design. This Bézier parametrization framework is being presented at the 2022 AIAA Aviation
Conference in Chicago, IL under the title "A Parametrization Framework for Multi-Element Airfoil Systems Using Bézier
curves."


Motivation
==========
The creation of this package was motivated by a research aircraft application: the aerodynamic design of a
propulsion-airframe-integrated commercial transport aircraft. The cross-section of a wing or fuselage with integrated
propulsors can be represented, with some sacrifice in fidelity, as a quasi-2D multi-element airfoil system. This
multi-element airfoil system is comprised of a main airfoil (either the fuselage or main airfoil element), a hub
airfoil (representing the cross-section of an axisymmetric hub), and a nacelle airfoil (representing the cross-section
of an axisymmetric nacelle).

By using a well-defined parametrization framework, this airfoil system can be morphed or deformed in a variety of
ways simply by changing the value of any of the input parameters. These parameters are represented by
``pymead.core.param.Param`` objects in this framework. Defining the airfoil system in this way provides an
intuitive I/O interface for shape optimization or parametric sweeps.

In `pymead`, airfoils are comprised of a set of connected, arbitrary-order Bézier curves. Because Bézier curves have
the property that they always pass through their starting and ending control points, Bézier curve "joints" can be used
to force the airfoil surface to pass through a particular point in space. `pymead` forces all Bézier curve joints
within an airfoil to be :math:`G^0`, :math:`G^1`, and :math:`G^2` continuous, which is useful in general for surface
smoothness and in particular for computational fluid dynamics (CFD) packages where a discontinuity in the curvature
value at a point can cause undesired flow properties or even unconverged results.


Applications
============
It is the hope of the author that pymead is sufficiently flexible to be used for airfoil applications of
varying complexities, from simple, single-airfoil design to high-fidelity, multi-element airfoil shape optimization.
Other common multi-element airfoil systems, such as the high-lift configuration on an aircraft, are also target
applications for this software package.

One utility provided in this software package which may be useful in the start-up phase of airfoil design is
``pymead.utils.airfoil_matching.match_airfoil()``. This modules allows the matching of a particular parametrization
to any public airfoil geometry at [airfoiltools.com](http://airfoiltools.com/) using the gradient-based "SLSQP"
optimizer.


Acknowledgments
===============
This work was supported by NASA under award number 80NSSC19M0125 as part of the Center for High-Efficiency Electrical
Technologies for Aircraft (CHEETA). Logo courtesy of [NASA](https://www.nasa.gov/).


Contact Information
===================
**Author**: Matthew G Lauer

**Email**: mlauer2015@gmail.com


Version Notes
=============

1.1.1
-----
- Corrections to README.md for PyPi long project description (images not showing properly)

1.1.0
-----
- Made corrections on BaseAirfoilParams and AnchorPoint Args domains
- Added support for zero-curvature anchor points
  using 180-degree curvature control arm angles (or 90-degree curvature control arm angles for the leading edge)
- Added support for sharp-juncture anchor points with :math:`R=0` or :math:`R_{LE}=0`. Adding multiple consecutive sharp
  juncture anchor points creates line segments. Adding sharp-juncture anchor points violates the principle of slope and
  curvature continuity, but may be useful in some cases.

"""
import os

from PyQt5.QtCore import QCoreApplication, QSettings

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(BASE_DIR, "data")
RESOURCE_DIR = os.path.join(BASE_DIR, "resources")
EXAMPLES_DIR = os.path.join(BASE_DIR, "examples")
ICON_DIR = os.path.join(BASE_DIR, "icons")
PLUGINS_DIR = os.path.join(BASE_DIR, "plugins")
INCLUDE_FILES = [os.path.join(BASE_DIR, "core", "symmetry.py")]
GUI_DEFAULTS_DIR = os.path.join(BASE_DIR, "gui", "gui_settings", "defaults")
GUI_SETTINGS_DIR = os.path.join(BASE_DIR, "gui", "gui_settings")
GUI_THEMES_DIR = os.path.join(BASE_DIR, "gui", "gui_settings", "themes")
GUI_DEFAULT_AIRFOIL_DIR = os.path.join(BASE_DIR, "gui", "default_airfoil")
GUI_DIALOG_WIDGETS_DIR = os.path.join(BASE_DIR, "gui", "dialog_widgets")
TEST_DIR = os.path.join(BASE_DIR, "tests")

QCoreApplication.setOrganizationName("mlaero")
QCoreApplication.setApplicationName("pymead")

q_settings = QSettings()


class DependencyNotFoundError(Exception):
    pass



import pandas as pd
from PyQt5.QtWidgets import QTextEdit

from pymead.core.geometry_collection import GeometryCollection
from pymead.gui.dialogs import PymeadDialog


class AirfoilStatistics:
    def __init__(self, geo_col: GeometryCollection):
        self.geo_col = geo_col
        self.df = None
        self.data = None
        self.col_spaces = None
        self.generate_data()

    def generate_data(self):
        cols = ["Max. t/c", "x/c @ Max. t/c", "Area", "Self-Intersecting"]
        index = []
        self.data = {k: [] for k in cols}
        for airfoil_name, airfoil in self.geo_col.container()["airfoils"].items():
            thickness_data = airfoil.compute_thickness()
            self.data[cols[0]].append(thickness_data["t/c_max"])
            self.data[cols[1]].append(thickness_data["t/c_max_x/c_loc"])
            self.data[cols[2]].append(airfoil.compute_area())
            self.data[cols[3]].append(int(airfoil.check_self_intersection()))
            index.append(airfoil_name)
        self.df = pd.DataFrame(data=self.data, index=index)

    def convert_to_html(self, **kwargs):
        """Convenience function for Pandas DataFrame conversion to HTML"""
        return self.df.to_html(**kwargs)

    def convert_to_latex(self, **kwargs):
        """Convenience function for Pandas DataFrame conversion to LaTeX"""
        return self.df.to_latex(**kwargs)

    def generate_text_edit_widget(self, parent, **html_kwargs):
        text_edit = QTextEdit(parent)
        html = self.convert_to_html(**html_kwargs)
        text_edit.setHtml(html)
        text_edit.setMinimumWidth(500)
        text_edit.setReadOnly(True)
        return text_edit


class AirfoilStatisticsDialog(PymeadDialog):
    def __init__(self, parent, airfoil_stats: AirfoilStatistics, theme: dict):
        self.stats_widget = airfoil_stats.generate_text_edit_widget(parent=parent, float_format="{:.8f}".format)
        super().__init__(parent=parent, window_title="Airfoil Statistics", widget=self.stats_widget,
                         theme=theme)



"""
Analysis module for `pymead`. Contains a partially complete Python API to both XFOIL and the MSES software suite. Also
contains a linear strength vortex panel method for a fast, purely inviscid airfoil flow solution sped up using Numba.
"""


import pyqtgraph as pg


class OptAirfoilGraph:
    def __init__(self, pen=None, size: tuple = (1000, 300), background_color: str = 'w'):
        pg.setConfigOptions(antialias=True)

        if pen is None:
            pen = pg.mkPen(color='cornflowerblue', width=2)

        self.w = pg.GraphicsLayoutWidget(show=True, size=size)
        # self.w.setWindowTitle('Airfoil')
        self.w.setBackground(background_color)
        self.v = self.w.addPlot(pen=pen)
        self.v.setAspectLocked()
        self.v.setLabel(axis='bottom', text='x')
        self.v.setLabel(axis='left', text='y')
        self.legend = self.v.addLegend(offset=(300, 20))
        # self.v.setAspectLocked()

    def set_background(self, background_color: str):
        self.w.setBackground(background_color)



from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import Qt


class SideGrip(QWidget):
    def __init__(self, parent, edge):
        super().__init__(parent)
        if edge == Qt.LeftEdge:
            self.setCursor(Qt.SizeHorCursor)
            self.resizeFunc = self.resizeLeft
        elif edge == Qt.TopEdge:
            self.setCursor(Qt.SizeVerCursor)
            self.resizeFunc = self.resizeTop
        elif edge == Qt.RightEdge:
            self.setCursor(Qt.SizeHorCursor)
            self.resizeFunc = self.resizeRight
        else:
            self.setCursor(Qt.SizeVerCursor)
            self.resizeFunc = self.resizeBottom
        self.mousePos = None

    def resizeLeft(self, delta):
        window = self.window()
        width = max(window.minimumWidth(), window.width() - delta.x())
        geo = window.geometry()
        geo.setLeft(geo.right() - width)
        window.setGeometry(geo)

    def resizeTop(self, delta):
        window = self.window()
        height = max(window.minimumHeight(), window.height() - delta.y())
        geo = window.geometry()
        geo.setTop(geo.bottom() - height)
        window.setGeometry(geo)

    def resizeRight(self, delta):
        window = self.window()
        width = max(window.minimumWidth(), window.width() + delta.x())
        window.resize(width, window.height())

    def resizeBottom(self, delta):
        window = self.window()
        height = max(window.minimumHeight(), window.height() + delta.y())
        window.resize(window.width(), height)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.mousePos = event.pos()

    def mouseMoveEvent(self, event):
        if self.mousePos is not None:
            delta = event.pos() - self.mousePos
            self.resizeFunc(delta)

    def mouseReleaseEvent(self, event):
        self.mousePos = None



import pyqtgraph as pg
from pyqtgraph.GraphicsScene.mouseEvents import HoverEvent
import shapely.geometry
from PyQt5.QtGui import QPicture, QPainter, QPolygonF, QPainterPath
from PyQt5.QtCore import QPointF, QRectF, pyqtSignal


# Create a subclass of GraphicsObject.
# The only required methods are paint() and boundingRect()
# (see QGraphicsItem documentation)
class PolygonItem(pg.GraphicsObject):
    sigPolyEnter = pyqtSignal(object, float, float)
    sigPolyExit = pyqtSignal(object)

    def __init__(self, data, airfoil, gui_obj, pen=None, brush=None):
        pg.GraphicsObject.__init__(self)
        self.data = data
        self.pen = pen
        self.brush = brush
        if self.pen is None:
            self.pen = pg.mkPen(255, 255, 255, 0)
            self.brush = pg.mkBrush(66, 233, 245, 50)
        self.picture = None
        self.polygon = QPolygonF()
        self.gui_obj = gui_obj
        self.airfoil = airfoil
        self.generatePicture()

    def update_polygon(self):
        """
        Updates the polygon based on a new set of vertices (contained in ``self.data``)
        """
        if len(self.polygon) != 0:
            self.polygon.clear()
        for row in self.data:
            self.polygon.append(QPointF(row[0], row[1]))

    def generatePicture(self):
        """
        Pre-computes the QPicture for speed.
        """
        self.picture = QPicture()
        p = QPainter(self.picture)
        p.setPen(self.pen)
        p.setBrush(self.brush)
        self.update_polygon()
        p.drawPolygon(self.polygon)
        p.end()
        if self.gui_obj.permanent_widget.inviscid_cl_combo.currentText() == self.airfoil.name():
            self.gui_obj.single_airfoil_inviscid_analysis(plot_cp=False)

    def shape(self):
        """
        Defines the shape of the polygon as a ``QPainterPath`` so that the hover detection is more accurate. Note that
        by default, Qt simply uses the shape returned by ``boundingRect()`` if this method is not overridden.

        Returns
        =======
        QPainterPath
            Painter path for the polygon
        """
        path = QPainterPath()
        path.addPolygon(self.polygon)
        path.closeSubpath()
        return path

    def paint(self, p, *args):
        p.drawPicture(0, 0, self.picture)

    def boundingRect(self):
        # boundingRect _must_ indicate the entire area that will be drawn on
        # or else we will get artifacts and possibly crashing.
        # (in this case, QPicture does all the work of computing the bounding rect for us)
        return QRectF(self.picture.boundingRect())

    def getCentroid(self):
        """
        Gets the centroid of the polygon using shapely

        Returns
        =======
        typing.List[float]
            List of [x,y] representing the centroid
        """
        shapely_poly = shapely.geometry.polygon.Polygon(shell=self.data)
        centroid = shapely_poly.centroid.xy
        return [centroid[0][0], centroid[1][0]]

    def hoverEvent(self, ev: HoverEvent):
        """
        Emits custom signals sigPolyEnter and sigPolyExit on airfoil polygon enter and exit

        Parameters
        ==========
        ev: pg.GraphicsScene.mouseEvents.HoverEvent
            Event emitted by pyqtgraph whenever the mouse hovers over the polygon defined by ``shape()``
        """
        centroid = self.getCentroid()
        if ev.isExit():
            self.sigPolyExit.emit(self.airfoil)
        elif ev.isEnter():
            self.sigPolyEnter.emit(self.airfoil, centroid[0], centroid[1])


if __name__ == '__main__':
    data_ = [[0, 0], [0.1, 0.1], [0.2, 0.3], [-1, 1], [0, 0]]
    item = PolygonItem(data_, "A0")
    plt = pg.plot()
    plt.addItem(item)
    plt.setWindowTitle('pyqtgraph example: customGraphicsItem')
    pg.exec()



from PyQt5.QtWidgets import QDoubleSpinBox
from PyQt5.QtGui import QKeyEvent
from PyQt5.QtCore import Qt
import numpy as np


class InftyDoubleSpinBox(QDoubleSpinBox):
    """
    From StackOverflow answer https://stackoverflow.com/a/57411410
    """
    def __init__(self, lower: bool):
        super(QDoubleSpinBox, self).__init__()

        self.setMinimum(-np.inf)
        self.setMaximum(np.inf)

        self.lower = lower

    def keyPressEvent(self, e: QKeyEvent):

        if e.key() == Qt.Key_Home:
            if self.lower:
                self.setValue(self.minimum())
            else:
                self.setValue(self.maximum())
        else:
            super(QDoubleSpinBox, self).keyPressEvent(e)
